import pytest


@pytest.mark.parametrize("length", [4], scope="module")
def test_nodes(in_memory_networkx_kg, _get_nodes):
    in_memory_networkx_kg.add_nodes(_get_nodes)
    networkx_kg = in_memory_networkx_kg.get_kg()
    expected_nodes = [
        (
            "p1",
            {
                "node_label": "protein",
                "score": 4.0,
                "name": "StringProperty1",
                "taxon": 9606,
                "genes": ["gene1", "gene2"],
                "id": "p1",
                "preferred_id": "uniprot",
            },
        ),
        (
            "p2",
            {
                "node_label": "protein",
                "score": 2.0,
                "name": "StringProperty1",
                "taxon": 9606,
                "genes": ["gene1", "gene2"],
                "id": "p2",
                "preferred_id": "uniprot",
            },
        ),
        (
            "p3",
            {
                "node_label": "protein",
                "score": 1.3333333333333333,
                "name": "StringProperty1",
                "taxon": 9606,
                "genes": ["gene1", "gene2"],
                "id": "p3",
                "preferred_id": "uniprot",
            },
        ),
        (
            "p4",
            {
                "node_label": "protein",
                "score": 1.0,
                "name": "StringProperty1",
                "taxon": 9606,
                "genes": ["gene1", "gene2"],
                "id": "p4",
                "preferred_id": "uniprot",
            },
        ),
        (
            "m1",
            {
                "node_label": "microRNA",
                "name": "StringProperty1",
                "taxon": 9606,
                "id": "m1",
                "preferred_id": "mirbase",
            },
        ),
        (
            "m2",
            {
                "node_label": "microRNA",
                "name": "StringProperty1",
                "taxon": 9606,
                "id": "m2",
                "preferred_id": "mirbase",
            },
        ),
        (
            "m3",
            {
                "node_label": "microRNA",
                "name": "StringProperty1",
                "taxon": 9606,
                "id": "m3",
                "preferred_id": "mirbase",
            },
        ),
        (
            "m4",
            {
                "node_label": "microRNA",
                "name": "StringProperty1",
                "taxon": 9606,
                "id": "m4",
                "preferred_id": "mirbase",
            },
        ),
    ]
    assert list(networkx_kg.nodes(data=True)) == expected_nodes


@pytest.mark.parametrize("length", [4], scope="module")
def test_nodes_gen(in_memory_networkx_kg, _get_nodes):
    def node_gen():
        for node in _get_nodes:
            yield node

    in_memory_networkx_kg.add_nodes(node_gen())
    networkx_kg = in_memory_networkx_kg.get_kg()
    assert len(networkx_kg.nodes(data=True)) == 8


@pytest.mark.parametrize("length", [4], scope="module")
def test_duplicates(in_memory_networkx_kg, _get_nodes):
    nodes = _get_nodes + _get_nodes
    in_memory_networkx_kg.add_nodes(nodes)
    networkx_kg = in_memory_networkx_kg.get_kg()
    protein_nodes = [node for node in networkx_kg.nodes(data=True) if node[1]["node_label"] == "protein"]
    assert len(protein_nodes) == 4


@pytest.mark.parametrize("length", [8], scope="module")
def test_two_step_add(in_memory_networkx_kg, _get_nodes):
    in_memory_networkx_kg.add_nodes(_get_nodes[:4])
    in_memory_networkx_kg.add_nodes(_get_nodes[4:])
    networkx_kg = in_memory_networkx_kg.get_kg()
    protein_nodes = [node for node in networkx_kg.nodes(data=True) if node[1]["node_label"] == "protein"]
    assert len(protein_nodes) == 8


@pytest.mark.parametrize("length", [4], scope="module")
def test_edges(in_memory_networkx_kg, _get_edges):
    in_memory_networkx_kg.add_edges(_get_edges)
    networkx_kg = in_memory_networkx_kg.get_kg()
    expected_edges = [
        (
            "p0",
            "p1",
            {
                "relationship_id": "prel0",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "p1",
            "p2",
            {
                "relationship_id": "prel1",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "p2",
            "p3",
            {
                "relationship_id": "prel2",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "p3",
            "p4",
            {
                "relationship_id": "prel3",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "m0",
            "p1",
            {
                "relationship_id": "mrel0",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
        (
            "m1",
            "p2",
            {
                "relationship_id": "mrel1",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
        (
            "m2",
            "p3",
            {
                "relationship_id": "mrel2",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
        (
            "m3",
            "p4",
            {
                "relationship_id": "mrel3",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
    ]
    assert list(networkx_kg.edges(data=True)) == expected_edges


@pytest.mark.parametrize("length", [4], scope="module")
def test_edges_gen(in_memory_networkx_kg, _get_edges):
    def edge_gen():
        for edge in _get_edges:
            yield edge

    in_memory_networkx_kg.add_edges(edge_gen())
    networkx_kg = in_memory_networkx_kg.get_kg()
    expected_edges = [
        (
            "p0",
            "p1",
            {
                "relationship_id": "prel0",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "p1",
            "p2",
            {
                "relationship_id": "prel1",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "p2",
            "p3",
            {
                "relationship_id": "prel2",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "p3",
            "p4",
            {
                "relationship_id": "prel3",
                "relationship_label": "PERTURBED_IN_DISEASE",
                "residue": "T253",
                "level": 4,
            },
        ),
        (
            "m0",
            "p1",
            {
                "relationship_id": "mrel0",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
        (
            "m1",
            "p2",
            {
                "relationship_id": "mrel1",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
        (
            "m2",
            "p3",
            {
                "relationship_id": "mrel2",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
        (
            "m3",
            "p4",
            {
                "relationship_id": "mrel3",
                "relationship_label": "Is_Mutated_In",
                "site": "3-UTR",
                "confidence": 1,
            },
        ),
    ]
    assert list(networkx_kg.edges(data=True)) == expected_edges


@pytest.mark.parametrize("length", [4], scope="module")
def test_rel_as_nodes(in_memory_networkx_kg, _get_rel_as_nodes):
    in_memory_networkx_kg.add_nodes(_get_rel_as_nodes)
    networkx_kg = in_memory_networkx_kg.get_kg()
    expected_edges = [
        (
            "i1",
            "p1",
            {"relationship_id": None, "relationship_label": "IS_SOURCE_OF"},
        ),
        (
            "i1",
            "p3",
            {"relationship_id": None, "relationship_label": "IS_TARGET_OF"},
        ),
        (
            "i2",
            "p2",
            {"relationship_id": None, "relationship_label": "IS_SOURCE_OF"},
        ),
        (
            "i2",
            "p4",
            {"relationship_id": None, "relationship_label": "IS_TARGET_OF"},
        ),
        (
            "i3",
            "p3",
            {"relationship_id": None, "relationship_label": "IS_SOURCE_OF"},
        ),
        (
            "i3",
            "p5",
            {"relationship_id": None, "relationship_label": "IS_TARGET_OF"},
        ),
        (
            "i4",
            "p4",
            {"relationship_id": None, "relationship_label": "IS_SOURCE_OF"},
        ),
        (
            "i0",
            "p2",
            {"relationship_id": None, "relationship_label": "IS_TARGET_OF"},
        ),
    ]
    assert list(networkx_kg.edges(data=True)) == expected_edges



import pytest

from hypothesis import (
    given,
    strategies as st,
)

from biocypher._create import BioCypherEdge, BioCypherNode, BioCypherRelAsNode


@given(st.builds(BioCypherNode))
def test_node(node):
    assert isinstance(node.get_id(), str)
    assert isinstance(node.get_label(), str)
    assert isinstance(node.get_properties(), dict)
    assert isinstance(node.get_dict(), dict)

    assert "id" in node.get_properties().keys()


@given(st.builds(BioCypherEdge))
def test_edge(edge):
    assert isinstance(edge.get_id(), str) or edge.get_id() is None
    assert isinstance(edge.get_source_id(), str)
    assert isinstance(edge.get_target_id(), str)
    assert isinstance(edge.get_label(), str)
    assert isinstance(edge.get_properties(), dict)
    assert isinstance(edge.get_dict(), dict)


@given(st.builds(BioCypherRelAsNode))
def test_rel_as_node(rel_as_node):
    assert isinstance(rel_as_node.get_node(), BioCypherNode)
    assert isinstance(rel_as_node.get_source_edge(), BioCypherEdge)
    assert isinstance(rel_as_node.get_target_edge(), BioCypherEdge)


def test_rel_as_node_invalid_node():
    with pytest.raises(TypeError):
        BioCypherRelAsNode("str", 1, 2.5122)



import pytest

from biocypher._create import BioCypherEdge, BioCypherNode, BioCypherRelAsNode


@pytest.fixture(scope="function")
def _get_nodes(length: int) -> list:
    nodes = []
    for i in range(length):
        bnp = BioCypherNode(
            node_id=f"p{i + 1}",
            node_label="protein",
            preferred_id="uniprot",
            properties={
                "score": 4 / (i + 1),
                "name": "StringProperty1",
                "taxon": 9606,
                "genes": ["gene1", "gene2"],
            },
        )
        nodes.append(bnp)
        bnm = BioCypherNode(
            node_id=f"m{i + 1}",
            node_label="microRNA",
            preferred_id="mirbase",
            properties={
                "name": "StringProperty1",
                "taxon": 9606,
            },
        )
        nodes.append(bnm)

    return nodes


@pytest.fixture(scope="function")
def _get_nodes_non_compliant_names(length: int) -> list:
    nodes = []
    for i in range(length):
        bnp = BioCypherNode(
            node_id=f"p{i + 1}",
            node_label="Patient (person)",
            preferred_id="snomedct",
            properties={},
        )
        nodes.append(bnp)
        bnm = BioCypherNode(
            node_id=f"m{i + 1}",
            node_label="1$He524ll<o wor.ld <(",
            preferred_id="snomedct",
            properties={},
        )
        nodes.append(bnm)

    return nodes


@pytest.fixture(scope="function")
def _get_edges(length):
    edges = []
    for i in range(length):
        e1 = BioCypherEdge(
            relationship_id=f"prel{i}",
            source_id=f"p{i}",
            target_id=f"p{i + 1}",
            relationship_label="PERTURBED_IN_DISEASE",
            properties={
                "residue": "T253",
                "level": 4,
            },
            # we suppose the verb-form relationship label is created by
            # translation functionality in translate.py
        )
        edges.append(e1)
        e2 = BioCypherEdge(
            relationship_id=f"mrel{i}",
            source_id=f"m{i}",
            target_id=f"p{i + 1}",
            relationship_label="Is_Mutated_In",
            properties={
                "site": "3-UTR",
                "confidence": 1,
            },
            # we suppose the verb-form relationship label is created by
            # translation functionality in translate.py
        )
        edges.append(e2)
    return edges


@pytest.fixture(scope="function")
def _get_edges_non_compliant_names(length):
    edges = []
    for i in range(length):
        e1 = BioCypherEdge(
            relationship_id=f"prel{i}",
            source_id=f"p{i}",
            target_id=f"p{i + 1}",
            relationship_label="(Compliant) edge",
            properties={},
        )
        edges.append(e1)
        e2 = BioCypherEdge(
            relationship_id=f"mrel{i}",
            source_id=f"m{i}",
            target_id=f"p{i + 1}",
            relationship_label="42Is_Mutated_In",
            properties={
                "site": "3-UTR",
                "confidence": 1,
            },
            # we suppose the verb-form relationship label is created by
            # translation functionality in translate.py
        )
        edges.append(e2)
    return edges


@pytest.fixture(scope="function")
def _get_rel_as_nodes(length):
    rels = []
    for i in range(length):
        n = BioCypherNode(
            node_id=f"i{i + 1}",
            node_label="post translational interaction",
            properties={
                "directed": True,
                "effect": -1,
            },
        )
        e1 = BioCypherEdge(
            source_id=f"i{i + 1}",
            target_id=f"p{i + 1}",
            relationship_label="IS_SOURCE_OF",
        )
        e2 = BioCypherEdge(
            source_id=f"i{i}",
            target_id=f"p{i + 2}",
            relationship_label="IS_TARGET_OF",
        )
        rels.append(BioCypherRelAsNode(n, e1, e2))
    return rels


# Immunology data: TCR alpha and beta chains + epitopes


# TCR alpha chain test nodes
@pytest.fixture()
def tra_nodes():
    """Return a list of TCR alpha chain nodes for testing."""
    return [
        BioCypherNode(
            node_id="tra:CAVRWGGKLSF",
            node_label="tra sequence",
            preferred_id="tra:CAVRWGGKLSF",
            properties={
                "junction_aa": "CAVRWGGKLSF",
                "chain_1_type": "tra",
                "chain_1_organism": "HomoSapiens",
                "chain_1_v_gene": "TRAV3*01",
                "chain_1_j_gene": "TRAJ20*01",
            },
        ),
        BioCypherNode(
            node_id="tra:CAGLLPGGGADGLTF",
            node_label="tra sequence",
            preferred_id="tra:CAGLLPGGGADGLTF",
            properties={
                "junction_aa": "CAGLLPGGGADGLTF",
                "chain_1_type": "tra",
                "chain_1_organism": "HomoSapiens",
                "chain_1_v_gene": "TRAV25*01",
                "chain_1_j_gene": "TRAJ45*01",
            },
        ),
        BioCypherNode(
            node_id="tra:CAVDNNNDMRF",
            node_label="tra sequence",
            preferred_id="tra:CAVDNNNDMRF",
            properties={
                "junction_aa": "CAVDNNNDMRF",
                "chain_1_type": "tra",
                "chain_1_v_gene": "TRAV12-2",
                "chain_1_j_gene": "TRAJ24",
            },
        ),
    ]


# TCR beta chain test nodes
@pytest.fixture()
def trb_nodes():
    """Return a list of TCR beta chain nodes for testing."""
    return [
        BioCypherNode(
            node_id="trb:CASSEGGVETQYF",
            node_label="trb sequence",
            preferred_id="trb:CASSEGGVETQYF",
            properties={
                "junction_aa": "CASSEGGVETQYF",
                "chain_1_type": "trb",
                "chain_1_organism": "HomoSapiens",
                "chain_1_v_gene": "TRBV13*01",
                "chain_1_j_gene": "TRBJ2-5*01",
            },
        ),
        BioCypherNode(
            node_id="trb:CASSSRGGQETQYF",
            node_label="trb sequence",
            preferred_id="trb:CASSSRGGQETQYF",
            properties={
                "junction_aa": "CASSSRGGQETQYF",
                "chain_1_type": "trb",
                "chain_1_organism": "HomoSapiens",
                "chain_1_v_gene": "TRBV7-3*01",
                "chain_1_j_gene": "TRBJ2-5*01",
            },
        ),
        BioCypherNode(
            node_id="trb:CASSPRGDSGNTIYF",
            node_label="trb sequence",
            preferred_id="trb:CASSPRGDSGNTIYF",
            properties={
                "junction_aa": "CASSPRGDSGNTIYF",
                "chain_1_type": "trb",
                "chain_1_v_gene": "TRBV7-9",
                "chain_1_j_gene": "TRBJ2-2",
            },
        ),
    ]


# Epitope test nodes
@pytest.fixture()
def epitope_nodes():
    """Return a list of epitope nodes for testing."""
    return [
        BioCypherNode(
            node_id="epitope:NLVPMVATV",
            node_label="epitope",
            preferred_id="epitope:NLVPMVATV",
            properties={
                "antigen_name": "pp65",
                "antigen_organism": "CMV",
                "MHC_class": "MHCI",
            },
        ),
        BioCypherNode(
            node_id="epitope:KLGGALQAK",
            node_label="epitope",
            preferred_id="epitope:KLGGALQAK",
            properties={
                "antigen_name": "IE1",
                "antigen_organism": "CMV",
                "MHC_class": "MHCI",
            },
        ),
        BioCypherNode(
            node_id="epitope:GILGFVFTL",
            node_label="epitope",
            preferred_id="epitope:GILGFVFTL",
            properties={
                "antigen_name": "M",
                "antigen_organism": "InfluenzaA",
                "MHC_class": "MHCI",
            },
        ),
    ]


# TCR pairing edges
@pytest.fixture()
def tcr_pair_edges():
    """Return a list of TCR alpha-beta pairing edges for testing."""
    return [
        BioCypherEdge(
            source_id="tra:CAVRWGGKLSF",
            target_id="trb:CASSEGGVETQYF",
            relationship_id="tra:CAVRWGGKLSF-trb:CASSEGGVETQYF",
            relationship_label="alpha sequence to beta sequence association",
        ),
        BioCypherEdge(
            source_id="tra:CAGLLPGGGADGLTF",
            target_id="trb:CASSSRGGQETQYF",
            relationship_id="tra:CAGLLPGGGADGLTF-trb:CASSSRGGQETQYF",
            relationship_label="alpha sequence to beta sequence association",
        ),
        BioCypherEdge(
            source_id="tra:CAVDNNNDMRF",
            target_id="trb:CASSPRGDSGNTIYF",
            relationship_id="tra:CAVDNNNDMRF-trb:CASSPRGDSGNTIYF",
            relationship_label="alpha sequence to beta sequence association",
        ),
    ]


# TCR-epitope association edges
@pytest.fixture()
def tcr_epitope_edges():
    """Return a list of TCR-to-epitope binding edges for testing."""
    return [
        BioCypherEdge(
            source_id="tra:CAVRWGGKLSF",
            target_id="epitope:NLVPMVATV",
            relationship_id="bind1",
            relationship_label="t cell receptor sequence to epitope association",
        ),
        BioCypherEdge(
            source_id="tra:CAVRWGGKLSF",
            target_id="epitope:KLGGALQAK",
            relationship_id="bind2",
            relationship_label="t cell receptor sequence to epitope association",
        ),
        BioCypherEdge(
            source_id="trb:CASSPRGDSGNTIYF",
            target_id="epitope:NLVPMVATV",
            relationship_id="bind3",
            relationship_label="t cell receptor sequence to epitope association",
        ),
        BioCypherEdge(
            source_id="tra:CAVDNNNDMRF",
            target_id="epitope:GILGFVFTL",
            relationship_id="bind4",
            relationship_label="t cell receptor sequence to epitope association",
        ),
        BioCypherEdge(
            source_id="tra:CAVDNNNDMRF",
            target_id="epitope:NLVPMVATV",
            relationship_id="bind5",
            relationship_label="t cell receptor sequence to epitope association",
        ),
    ]



import pytest

from biocypher._mapping import OntologyMapping
from biocypher._ontology import Ontology, OntologyAdapter


@pytest.fixture(scope="module")
def ontology_mapping():
    return OntologyMapping(config_file="biocypher/_config/test_schema_config.yaml")


@pytest.fixture(scope="module")
def simple_ontology_mapping():
    m = OntologyMapping()
    m.schema = {
        "accuracy": {},
    }
    return m


@pytest.fixture(scope="module")
def extended_ontology_mapping():
    return OntologyMapping(config_file="biocypher/_config/test_schema_config_extended.yaml")


@pytest.fixture(scope="module")
def disconnected_mapping():
    return OntologyMapping(config_file="biocypher/_config/test_schema_config_disconnected.yaml")


@pytest.fixture(scope="module")
def hybrid_ontology(extended_ontology_mapping):
    return Ontology(
        head_ontology={
            "url": "https://github.com/biolink/biolink-model/raw/v3.2.1/biolink-model.owl.ttl",
            "root_node": "entity",
        },
        ontology_mapping=extended_ontology_mapping,
        tail_ontologies={
            "so": {
                "url": "test/ontologies/so.owl",
                "head_join_node": "sequence variant",
                "tail_join_node": "sequence_variant",
            },
            "mondo": {
                "url": "test/ontologies/mondo.owl",
                "head_join_node": "disease",
                "tail_join_node": "human disease",
                "merge_nodes": False,
            },
        },
    )


@pytest.fixture(scope="module")
def simple_ontology(simple_ontology_mapping):
    return Ontology(
        head_ontology={
            "url": "test/ontologies/ontology1.ttl",
            "root_node": "Thing",
        },
        ontology_mapping=simple_ontology_mapping,
        tail_ontologies={
            "test": {
                "url": "test/ontologies/ontology2.ttl",
                "head_join_node": "entity",
                "tail_join_node": "EvaluationCriterion",
            },
        },
    )


@pytest.fixture(scope="module")
def biolink_adapter():
    return OntologyAdapter(
        "https://github.com/biolink/biolink-model/raw/v3.2.1/biolink-model.owl.ttl",
        "entity",
    )


@pytest.fixture(scope="module")
def so_adapter():
    return OntologyAdapter("test/ontologies/so.owl", "sequence_variant")


@pytest.fixture(scope="module")
def go_adapter():
    return OntologyAdapter("test/ontologies/go.owl", "molecular_function")


@pytest.fixture(scope="module")
def mondo_adapter():
    return OntologyAdapter("test/ontologies/mondo.owl", "disease")



import pytest

from biocypher._config import _read_yaml


def test_read_yaml():
    schema_config = _read_yaml("biocypher/_config/test_schema_config.yaml")

    assert "protein" in schema_config


def test_for_special_characters():
    with pytest.warns(UserWarning):
        _read_yaml("biocypher/_config/test_config.yaml")



import pytest

from biocypher.output.write.graph._networkx import _NetworkXWriter


@pytest.fixture(scope="function")
def bw_networkx(translator, deduplicator, tmp_path_session):
    bw_networkx = _NetworkXWriter(
        translator=translator,
        deduplicator=deduplicator,
        output_directory=tmp_path_session,
    )

    yield bw_networkx



import pytest

from biocypher.output.in_memory._networkx import NetworkxKG
from biocypher.output.in_memory._pandas import PandasKG
from biocypher.output.in_memory._airr import AirrKG


@pytest.fixture(scope="function")
def in_memory_pandas_kg(deduplicator):
    in_memory_kg = PandasKG(
        deduplicator=deduplicator,
    )

    yield in_memory_kg


@pytest.fixture(scope="function")
def in_memory_networkx_kg(deduplicator):
    in_memory_kg = NetworkxKG(
        deduplicator=deduplicator,
    )

    yield in_memory_kg


@pytest.fixture(scope="function")
def in_memory_airr_kg(deduplicator):
    in_memory_kg = AirrKG(
        deduplicator=deduplicator,
    )

    yield in_memory_kg



import os

import pytest

from biocypher.output.write.graph._rdf import _RDFWriter


@pytest.fixture(scope="function")
def bw_rdf(translator, deduplicator, tmp_path_session):
    bw_rdf = _RDFWriter(
        translator=translator,
        deduplicator=deduplicator,
        output_directory=tmp_path_session,
        file_format="xml",
        rdf_namespaces={},
        delimiter=",",
    )
    yield bw_rdf

    # teardown
    for f in os.listdir(tmp_path_session):
        os.remove(os.path.join(tmp_path_session, f))


@pytest.fixture(scope="function")
def bw_rdf_ttl(translator, deduplicator, tmp_path_session):
    """Fixture for RDF writer with ttl format."""
    bw_rdf_ttl = _RDFWriter(
        translator=translator,
        deduplicator=deduplicator,
        output_directory=tmp_path_session,
        file_format="ttl",  # Using ttl format directly
        rdf_namespaces={},
        delimiter=",",
    )
    yield bw_rdf_ttl

    # teardown
    for f in os.listdir(tmp_path_session):
        os.remove(os.path.join(tmp_path_session, f))



"""Unified BioCypher Workflow API for knowledge graph workflows.

This module provides a streamlined interface for creating and managing
knowledge graphs using the unified Graph representation, with optional
schema and ontology support. Designed for both agentic and deterministic
workflows.

## Design Philosophy

This API is designed with the following principles:
1. **Agentic-First**: Optimized for LLM agent workflows with computable functions
2. **Zero Dependencies**: Pure Python implementation for maximum compatibility
3. **Future-Proof**: Native BioCypher objects enable advanced agentic features
4. **Migration-Ready**: Wrapper methods provide compatibility with existing tools
5. **Progressive Validation**: Optional validation and deduplication with flexible modes

## Validation and Deduplication

Unlike the legacy BioCypher which enforces strict validation and deduplication,
this API provides **progressive validation** with three modes:

- **"none"** (default): No validation or deduplication - maximum flexibility for agents
- **"warn"**: Log warnings for schema violations and duplicates but continue processing
- **"strict"**: Enforce schema validation and deduplication - fail fast on violations

This approach allows:
- **Agents** to work with maximum flexibility (no validation overhead)
- **Deterministic workflows** to use validation when needed
- **Gradual migration** from legacy BioCypher (start with "none", add validation later)

## Future Migration Path

This module represents the future direction of BioCypher's in-memory graph
functionality. The plan is to:

1. **Phase 1** (Current): Keep separate from legacy code, provide compatibility wrappers
2. **Phase 2**: Replace legacy in-memory implementations (PandasKG, NetworkxKG)
3. **Phase 3**: Add advanced agentic features (computable functions, decision logging)
4. **Phase 4**: Integrate with main BioCypher class as unified interface

## Agentic Features (Future)

- Computable functions attached to nodes/edges
- Decision logging and reasoning traces
- Counterfactual inference capabilities
- MCP (Model Context Protocol) interface integration
- Local graph computation for agent workflows

"""

import json

from typing import Any

import yaml

from ._graph import Edge, Graph, HyperEdge, Node
from ._logger import logger


class BioCypherWorkflow:
    """Unified BioCypher interface for knowledge graph workflows.

    This class provides a clean, simple API for creating and managing
    knowledge graphs with optional schema and ontology support. Designed
    for both agentic and deterministic workflows.
    """

    def __init__(
        self,
        name: str = "workflow_graph",
        directed: bool = True,
        schema: dict[str, Any] | None = None,
        schema_file: str | None = None,
        head_ontology_url: str | None = None,
        validation_mode: str = "none",
        deduplication: bool = False,
    ):
        """Initialize the workflow with a unified graph.

        Args:
            name: Name of the knowledge graph
            directed: Whether the graph is directed (default: True)
            schema: Dictionary defining the knowledge graph schema
            schema_file: Path to YAML schema file
            head_ontology_url: URL to ontology file (defaults to Biolink model)
            validation_mode: Validation level ("none", "warn", "strict")
            deduplication: Whether to enable deduplication (default: False)
        """
        self.graph = Graph(name=name, directed=directed)
        self.name = name
        self.schema = schema
        self.schema_file = schema_file
        self.head_ontology_url = head_ontology_url
        self.validation_mode = validation_mode
        self.deduplication = deduplication

        # Track seen entities for deduplication
        self._seen_nodes = set()
        self._seen_edges = set()

        # Initialize schema if provided
        if schema_file:
            self._load_schema_from_file(schema_file)
        elif schema:
            self._load_schema_from_dict(schema)

    def _load_schema_from_file(self, schema_file: str) -> None:
        """Load schema from YAML file."""
        try:
            with open(schema_file, "r") as f:
                self.schema = yaml.safe_load(f)
            logger.info(f"Loaded schema from {schema_file}")
        except Exception as e:
            logger.warning(f"Could not load schema from {schema_file}: {e}")

    def _load_schema_from_dict(self, schema: dict[str, Any]) -> None:
        """Load schema from dictionary."""
        self.schema = schema
        logger.info("Loaded schema from dictionary")

    # ==================== NODE OPERATIONS ====================

    def add_node(self, node_id: str, node_type: str, **properties) -> bool:
        """Add a node to the knowledge graph.

        Args:
            node_id: Unique identifier for the node
            node_type: Type/category of the node
            **properties: Node properties as keyword arguments

        Returns:
            bool: True if node was added, False if it already exists

        Example:
            workflow.add_node("protein_1", "protein", name="TP53", function="tumor_suppressor")
        """
        # Check for duplicates if deduplication is enabled
        if self.deduplication:
            if node_id in self._seen_nodes:
                if self.validation_mode == "warn":
                    logger.warning(f"Duplicate node ID '{node_id}' detected")
                elif self.validation_mode == "strict":
                    raise ValueError(f"Duplicate node ID '{node_id}' not allowed in strict mode")
                return False
            self._seen_nodes.add(node_id)

        # Validate against schema if validation is enabled
        if self.validation_mode in ["warn", "strict"]:
            is_valid = self.validate_against_schema(node_type, properties)
            if not is_valid:
                if self.validation_mode == "strict":
                    raise ValueError(f"Node '{node_id}' of type '{node_type}' failed schema validation")
                elif self.validation_mode == "warn":
                    logger.warning(f"Node '{node_id}' of type '{node_type}' failed schema validation")

        # Try to add node to graph (Graph class handles its own deduplication)
        result = self.graph.add_node(node_id, node_type, properties)

        # If deduplication is enabled and we're tracking, update our tracking
        if self.deduplication and result:
            self._seen_nodes.add(node_id)

        return result

    def get_node(self, node_id: str) -> Node | None:
        """Get a node by ID.

        Args:
            node_id: Node identifier

        Returns:
            Node object or None if not found
        """
        return self.graph.get_node(node_id)

    def get_nodes(self, node_type: str | None = None) -> list[Node]:
        """Get all nodes, optionally filtered by type.

        Args:
            node_type: Optional filter by node type

        Returns:
            List of Node objects
        """
        return self.graph.get_nodes(node_type)

    def has_node(self, node_id: str) -> bool:
        """Check if a node exists.

        Args:
            node_id: Node identifier

        Returns:
            bool: True if node exists
        """
        return self.graph.has_node(node_id)

    def remove_node(self, node_id: str) -> bool:
        """Remove a node and all its connected edges.

        Args:
            node_id: Node identifier

        Returns:
            bool: True if node was removed, False if not found
        """
        return self.graph.remove_node(node_id)

    # ==================== EDGE OPERATIONS ====================

    def add_edge(self, edge_id: str, edge_type: str, source: str, target: str, **properties) -> bool:
        """Add an edge to the knowledge graph.

        Args:
            edge_id: Unique identifier for the edge
            edge_type: Type/category of the edge
            source: Source node ID
            target: Target node ID
            **properties: Edge properties as keyword arguments

        Returns:
            bool: True if edge was added, False if it already exists

        Example:
            workflow.add_edge("interaction_1", "interaction", "protein_1", "protein_2",
                          confidence=0.8, method="yeast_two_hybrid")
        """
        # Check for duplicates if deduplication is enabled
        if self.deduplication:
            edge_key = (edge_id, edge_type)
            if edge_key in self._seen_edges:
                if self.validation_mode == "warn":
                    logger.warning(f"Duplicate edge ID '{edge_id}' of type '{edge_type}' detected")
                elif self.validation_mode == "strict":
                    raise ValueError(f"Duplicate edge ID '{edge_id}' not allowed in strict mode")
                return False
            self._seen_edges.add(edge_key)

        # Validate against schema if validation is enabled
        if self.validation_mode in ["warn", "strict"]:
            is_valid = self.validate_against_schema(edge_type, properties)
            if not is_valid:
                if self.validation_mode == "strict":
                    raise ValueError(f"Edge '{edge_id}' of type '{edge_type}' failed schema validation")
                elif self.validation_mode == "warn":
                    logger.warning(f"Edge '{edge_id}' of type '{edge_type}' failed schema validation")

        # Try to add edge to graph (Graph class handles its own deduplication)
        result = self.graph.add_edge(edge_id, edge_type, source, target, properties)

        # If deduplication is enabled and we're tracking, update our tracking
        if self.deduplication and result:
            edge_key = (edge_id, edge_type)
            self._seen_edges.add(edge_key)

        return result

    def get_edge(self, edge_id: str) -> Edge | None:
        """Get an edge by ID.

        Args:
            edge_id: Edge identifier

        Returns:
            Edge object or None if not found
        """
        return self.graph.get_edge(edge_id)

    def get_edges(self, edge_type: str | None = None) -> list[Edge]:
        """Get all edges, optionally filtered by type.

        Args:
            edge_type: Optional filter by edge type

        Returns:
            List of Edge objects
        """
        return self.graph.get_edges(edge_type)

    def get_edges_between(self, source: str, target: str, edge_type: str | None = None) -> list[Edge]:
        """Get edges between two nodes.

        Args:
            source: Source node ID
            target: Target node ID
            edge_type: Optional filter by edge type

        Returns:
            List of Edge objects
        """
        return self.graph.get_edges_between(source, target, edge_type)

    def has_edge(self, edge_id: str) -> bool:
        """Check if an edge exists.

        Args:
            edge_id: Edge identifier

        Returns:
            bool: True if edge exists
        """
        return self.graph.has_edge(edge_id)

    def remove_edge(self, edge_id: str) -> bool:
        """Remove an edge from the graph.

        Args:
            edge_id: Edge identifier

        Returns:
            bool: True if edge was removed, False if not found
        """
        return self.graph.remove_edge(edge_id)

    # ==================== HYPEREDGE OPERATIONS ====================

    def add_hyperedge(self, hyperedge_id: str, hyperedge_type: str, nodes: set[str], **properties) -> bool:
        """Add a hyperedge connecting multiple nodes.

        Args:
            hyperedge_id: Unique identifier for the hyperedge
            hyperedge_type: Type/category of the hyperedge
            nodes: Set of node IDs to connect
            **properties: Hyperedge properties as keyword arguments

        Returns:
            bool: True if hyperedge was added, False if it already exists

        Example:
            workflow.add_hyperedge("complex_1", "protein_complex", {"protein_1", "protein_2", "protein_3"},
                               name="transcription_factor_complex")
        """
        return self.graph.add_hyperedge(hyperedge_id, hyperedge_type, nodes, properties)

    def get_hyperedge(self, hyperedge_id: str) -> HyperEdge | None:
        """Get a hyperedge by ID.

        Args:
            hyperedge_id: Hyperedge identifier

        Returns:
            HyperEdge object or None if not found
        """
        return self.graph.get_hyperedge(hyperedge_id)

    def get_hyperedges(self, hyperedge_type: str | None = None) -> list[HyperEdge]:
        """Get all hyperedges, optionally filtered by type.

        Args:
            hyperedge_type: Optional filter by hyperedge type

        Returns:
            List of HyperEdge objects
        """
        return self.graph.get_hyperedges(hyperedge_type)

    def has_hyperedge(self, hyperedge_id: str) -> bool:
        """Check if a hyperedge exists.

        Args:
            hyperedge_id: Hyperedge identifier

        Returns:
            bool: True if hyperedge exists
        """
        return self.graph.has_hyperedge(hyperedge_id)

    # ==================== GRAPH TRAVERSAL ====================

    def get_neighbors(self, node_id: str, direction: str = "both") -> set[str]:
        """Get neighboring nodes.

        Args:
            node_id: Node identifier
            direction: "in", "out", or "both"

        Returns:
            Set of neighboring node IDs
        """
        return self.graph.get_neighbors(node_id, direction)

    def get_connected_edges(self, node_id: str, direction: str = "both") -> list[Edge]:
        """Get edges connected to a node.

        Args:
            node_id: Node identifier
            direction: "in", "out", or "both"

        Returns:
            List of connected Edge objects
        """
        return self.graph.get_connected_edges(node_id, direction)

    def find_paths(self, source: str, target: str, max_length: int = 3) -> list[list[Edge]]:
        """Find all paths between two nodes.

        Args:
            source: Source node ID
            target: Target node ID
            max_length: Maximum path length

        Returns:
            List of paths, each path is a list of Edge objects
        """
        return self.graph.find_paths(source, target, max_length)

    # ==================== QUERY INTERFACE ====================

    def query_nodes(self, node_type: str | None = None) -> list[dict[str, Any]]:
        """Query nodes in the knowledge graph.

        Args:
            node_type: Optional filter by node type

        Returns:
            List of node dictionaries
        """
        nodes = self.graph.get_nodes(node_type)
        return [node.to_dict() for node in nodes]

    def query_edges(self, edge_type: str | None = None) -> list[dict[str, Any]]:
        """Query edges in the knowledge graph.

        Args:
            edge_type: Optional filter by edge type

        Returns:
            List of edge dictionaries
        """
        edges = self.graph.get_edges(edge_type)
        return [edge.to_dict() for edge in edges]

    def query_hyperedges(self, hyperedge_type: str | None = None) -> list[dict[str, Any]]:
        """Query hyperedges in the knowledge graph.

        Args:
            hyperedge_type: Optional filter by hyperedge type

        Returns:
            List of hyperedge dictionaries
        """
        hyperedges = self.graph.get_hyperedges(hyperedge_type)
        return [hyperedge.to_dict() for hyperedge in hyperedges]

    def find_connected_components(self, node_id: str, max_depth: int = 2) -> dict[str, Any]:
        """Find connected components around a node.

        Args:
            node_id: Starting node ID
            max_depth: Maximum depth to explore

        Returns:
            Dictionary with nodes and edges in the component
        """
        if not self.graph.has_node(node_id):
            return {"nodes": [], "edges": [], "hyperedges": []}

        # Collect nodes within max_depth
        component_nodes = {node_id}
        current_level = {node_id}

        for depth in range(max_depth):
            next_level = set()
            for node in current_level:
                neighbors = self.graph.get_neighbors(node)
                next_level.update(neighbors)
            current_level = next_level - component_nodes
            component_nodes.update(current_level)

            if not current_level:
                break

        # Get subgraph
        subgraph = self.graph.get_subgraph(component_nodes)

        return {
            "nodes": [node.to_dict() for node in subgraph.get_nodes()],
            "edges": [edge.to_dict() for edge in subgraph.get_edges()],
            "hyperedges": [hyperedge.to_dict() for hyperedge in subgraph.get_hyperedges()],
            "statistics": subgraph.get_statistics(),
        }

    # ==================== GRAPH ANALYSIS ====================

    def get_statistics(self) -> dict[str, Any]:
        """Get comprehensive graph statistics.

        Returns:
            Dictionary with graph statistics
        """
        return self.graph.get_statistics()

    def get_summary(self) -> dict[str, Any]:
        """Get a human-readable summary of the graph.

        Returns:
            Dictionary with graph summary
        """
        stats = self.graph.get_statistics()

        # Get top node types
        node_types = stats["node_types"]
        top_node_types = sorted(node_types.items(), key=lambda x: x[1], reverse=True)[:5]

        # Get top edge types
        edge_types = stats["edge_types"]
        top_edge_types = sorted(edge_types.items(), key=lambda x: x[1], reverse=True)[:5]

        return {
            "name": self.name,
            "total_nodes": stats["basic"]["nodes"],
            "total_edges": stats["basic"]["edges"],
            "total_hyperedges": stats["basic"]["hyperedges"],
            "top_node_types": top_node_types,
            "top_edge_types": top_edge_types,
            "connectivity": stats["connectivity"],
        }

    # ==================== SCHEMA AND ONTOLOGY SUPPORT ====================

    def get_schema(self) -> dict[str, Any] | None:
        """Get the current schema configuration.

        Returns:
            Dictionary representing the schema or None if no schema
        """
        return self.schema

    def export_schema(self, filepath: str) -> None:
        """Export the current schema to a YAML file.

        Args:
            filepath: Path to save the schema file
        """
        if self.schema:
            with open(filepath, "w") as f:
                yaml.dump(self.schema, f, default_flow_style=False)
            logger.info(f"Schema exported to {filepath}")
        else:
            logger.warning("No schema to export")

    def validate_against_schema(self, node_type: str, properties: dict[str, Any]) -> bool:
        """Validate node properties against schema (if available).

        Args:
            node_type: Type of node to validate
            properties: Properties to validate

        Returns:
            bool: True if valid, False otherwise
        """
        if not self.schema or node_type not in self.schema:
            return True  # No schema or type not in schema, assume valid

        schema_entry = self.schema[node_type]
        if "properties" not in schema_entry:
            return True  # No property constraints

        required_properties = schema_entry["properties"]

        # Check if all required properties are present and have correct types
        for prop_name, prop_type in required_properties.items():
            if prop_name not in properties:
                logger.warning(f"Missing required property '{prop_name}' for node type '{node_type}'")
                return False

            # Check property type
            actual_value = properties[prop_name]
            if not self._validate_property_type(actual_value, prop_type):
                logger.warning(
                    f"Property '{prop_name}' has wrong type. Expected {prop_type}, got {type(actual_value).__name__}"
                )
                return False

        return True

    def _validate_property_type(self, value: Any, expected_type: str) -> bool:
        """Validate that a property value matches the expected type.

        Args:
            value: The actual value
            expected_type: The expected type as string (e.g., 'str', 'int', 'float')

        Returns:
            bool: True if type matches, False otherwise
        """
        type_mapping = {
            "str": str,
            "int": int,
            "float": float,
            "bool": bool,
            "list": list,
            "dict": dict,
        }

        if expected_type not in type_mapping:
            return True  # Unknown type, assume valid

        expected_python_type = type_mapping[expected_type]
        return isinstance(value, expected_python_type)

    # ==================== SERIALIZATION ====================

    def to_json(self) -> str:
        """Export the knowledge graph to JSON format.

        Returns:
            JSON string representation of the graph
        """
        return self.graph.to_json()

    def from_json(self, json_data: str) -> None:
        """Import knowledge graph from JSON format.

        Args:
            json_data: JSON string containing graph data
        """
        data = json.loads(json_data)
        self.graph = Graph.from_dict(data)
        self.name = self.graph.name

    def save(self, filepath: str) -> None:
        """Save the graph to a file.

        Args:
            filepath: Path to save the graph
        """
        with open(filepath, "w") as f:
            f.write(self.to_json())
        logger.info(f"Graph saved to {filepath}")

    def load(self, filepath: str) -> None:
        """Load the graph from a file.

        Args:
            filepath: Path to load the graph from
        """
        with open(filepath, "r") as f:
            json_data = f.read()
        self.from_json(json_data)
        logger.info(f"Graph loaded from {filepath}")

    # ==================== UTILITY METHODS ====================

    def clear(self) -> None:
        """Clear all nodes and edges from the graph."""
        self.graph = Graph(name=self.name, directed=self.graph.directed)
        logger.info("Graph cleared")

    def copy(self) -> "BioCypherWorkflow":
        """Create a copy of the workflow and its graph.

        Returns:
            New BioCypherWorkflow instance
        """
        new_workflow = BioCypherWorkflow(
            name=self.name, directed=self.graph.directed, schema=self.schema, head_ontology_url=self.head_ontology_url
        )
        new_workflow.from_json(self.to_json())
        return new_workflow

    def get_graph(self) -> Graph:
        """Get the underlying Graph object.

        Returns:
            Graph object
        """
        return self.graph

    def __len__(self) -> int:
        """Return the number of nodes in the graph."""
        return len(self.graph)

    def __contains__(self, node_id: str) -> bool:
        """Check if a node exists in the graph."""
        return node_id in self.graph

    def __str__(self) -> str:
        """String representation of the workflow."""
        stats = self.get_statistics()
        return (
            f"BioCypherWorkflow(name='{self.name}', "
            f"nodes={stats['basic']['nodes']}, edges={stats['basic']['edges']}, "
            f"hyperedges={stats['basic']['hyperedges']})"
        )

    def __repr__(self) -> str:
        return self.__str__()

    # ==================== COMPATIBILITY WRAPPER METHODS ====================

    def to_networkx(self):
        """Convert to NetworkX graph for compatibility with existing tools.

        Returns:
            networkx.DiGraph: NetworkX representation of the graph

        Note:
            This method provides compatibility with existing NetworkX-based
            tools while maintaining the native BioCypher object structure.
            Future versions may use this as the primary backend.
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError("NetworkX is required for to_networkx() conversion. Install with: pip install networkx")

        g = nx.DiGraph() if self.graph.directed else nx.Graph()

        # Add nodes with properties
        for node in self.graph._nodes.values():
            attrs = node.properties.copy()
            attrs["node_type"] = node.type
            g.add_node(node.id, **attrs)

        # Add edges with properties
        for edge in self.graph._edges.values():
            attrs = edge.properties.copy()
            attrs["edge_type"] = edge.type
            g.add_edge(edge.source, edge.target, **attrs)

        return g

    def to_pandas(self):
        """Convert to Pandas DataFrames for compatibility with existing tools.

        Returns:
            dict[str, pd.DataFrame]: Dictionary of DataFrames, one per node/edge type

        Note:
            This method provides compatibility with existing Pandas-based
            tools while maintaining the native BioCypher object structure.
            Future versions may use this as the primary backend.
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("Pandas is required for to_pandas() conversion. Install with: pip install pandas")

        dfs = {}

        # Create node DataFrames by type
        for node_type, node_ids in self.graph._node_types.items():
            nodes = [self.graph._nodes[node_id] for node_id in node_ids]
            data = []
            for node in nodes:
                row = {"node_id": node.id, "node_type": node.type}
                row.update(node.properties)
                data.append(row)
            dfs[node_type] = pd.DataFrame(data)

        # Create edge DataFrames by type
        for edge_type, edge_ids in self.graph._edge_types.items():
            edges = [self.graph._edges[edge_id] for edge_id in edge_ids]
            data = []
            for edge in edges:
                row = {"edge_id": edge.id, "edge_type": edge.type, "source_id": edge.source, "target_id": edge.target}
                row.update(edge.properties)
                data.append(row)
            dfs[edge_type] = pd.DataFrame(data)

        return dfs


# Convenience function for quick workflow creation
def create_workflow(
    name: str = "knowledge_graph",
    directed: bool = True,
    schema: dict[str, Any] | None = None,
    schema_file: str | None = None,
    head_ontology_url: str | None = None,
    validation_mode: str = "none",
    deduplication: bool = False,
) -> BioCypherWorkflow:
    """Create a new knowledge graph workflow.

    Args:
        name: Name of the knowledge graph
        directed: Whether the graph is directed
        schema: Dictionary defining the knowledge graph schema
        schema_file: Path to YAML schema file
        head_ontology_url: URL to ontology file
        validation_mode: Validation level ("none", "warn", "strict")
        deduplication: Whether to enable deduplication

    Returns:
        BioCypherWorkflow instance
    """
    return BioCypherWorkflow(
        name=name,
        directed=directed,
        schema=schema,
        schema_file=schema_file,
        head_ontology_url=head_ontology_url,
        validation_mode=validation_mode,
        deduplication=deduplication,
    )



import os

import pytest

from biocypher.output.write.graph._arangodb import _ArangoDBBatchWriter


@pytest.fixture(scope="function")
def bw_arango(translator, deduplicator, tmp_path_session):
    bw_arango = _ArangoDBBatchWriter(
        translator=translator,
        deduplicator=deduplicator,
        output_directory=tmp_path_session,
        delimiter=",",
    )

    yield bw_arango

    # teardown
    for f in os.listdir(tmp_path_session):
        os.remove(os.path.join(tmp_path_session, f))



from biocypher import BioCypher
from tutorial.data_generator import Protein


def main():
    # Setup: create a list of proteins to be imported
    proteins = [Protein() for _ in range(10)]

    # Extract id, label, and property dictionary
    def node_generator():
        for protein in proteins:
            yield (
                protein.get_id(),
                protein.get_label(),
                protein.get_properties(),
            )

    # Create BioCypher driver
    bc = BioCypher(
        biocypher_config_path="tutorial/01_biocypher_config.yaml",
        schema_config_path="tutorial/01_schema_config.yaml",
    )
    # Run the import
    bc.write_nodes(node_generator())

    # Write command line call
    bc.write_import_call()


if __name__ == "__main__":
    main()



"""
Neo4j connection management and Cypher interface.

A wrapper around the Neo4j driver which handles the DBMS connection and
provides basic management methods. This module is only used when BioCypher
is configured for online mode with Neo4j.
"""

from __future__ import annotations

import contextlib
import itertools
import os
import re
import warnings

from typing import Literal

import appdirs
import yaml

from biocypher._logger import logger
from biocypher._misc import to_list

__all__ = ["CONFIG_FILES", "DEFAULT_CONFIG", "Neo4jDriver"]

# Try to import Neo4j driver, but don't fail if not available
try:
    import neo4j
    import neo4j.exceptions as neo4j_exc

    NEO4J_AVAILABLE = True
except ImportError:
    NEO4J_AVAILABLE = False
    neo4j = None
    neo4j_exc = None

CONFIG_FILES = Literal["neo4j.yaml", "neo4j.yml"]
DEFAULT_CONFIG = {
    "user": "neo4j",
    "passwd": "neo4j",
    "db": "neo4j",
    "uri": "neo4j://localhost:7687",
    "fetch_size": 1000,
    "raise_errors": False,
    "fallback_db": ("system", "neo4j"),
    "fallback_on": ("TransientError",),
}


def _to_tuple(value):
    """Ensure that value is a tuple."""
    return tuple(to_list(value))


def _to_set(value):
    """Ensure that value is a set."""
    return set(to_list(value))


def _if_none(*values):
    """Use the first item from values that is not None."""
    for v in values:
        if v is not None:
            return v
    return None


def _pretty_profile(d, lines=None, indent=0):
    """
    Pretty format a Neo4j profile dict.

    Takes Neo4j profile dictionary and an optional header as
    list and creates a list of strings to be printed.

    Args:
        d: Profile dictionary or list
        lines: Optional list to append to
        indent: Indentation level

    Returns:
        List of formatted strings
    """
    if lines is None:
        lines = []

    # ANSI color codes for terminal output
    OKBLUE = "\033[94m"
    WARNING = "\033[93m"
    ENDC = "\033[0m"

    # if more items, branch
    if d:
        if isinstance(d, list):
            for sd in d:
                _pretty_profile(sd, lines, indent)
        elif isinstance(d, dict):
            typ = d.pop("operatorType", None)
            if typ:
                lines.append(("\t" * indent) + "|" + "\t" + f"{OKBLUE}Step: {typ} {ENDC}")

            # buffer children
            chi = d.pop("children", None)

            for key, value in d.items():
                if key == "args":
                    _pretty_profile(value, lines, indent)
                # both are there for some reason, sometimes
                # both in the same process
                elif key == "Time" or key == "time":
                    lines.append(
                        ("\t" * indent) + "|" + "\t" + str(key) + ": " + f"{WARNING}{value:,}{ENDC}".replace(",", " ")
                    )
                else:
                    lines.append(("\t" * indent) + "|" + "\t" + str(key) + ": " + str(value))

            # now the children
            _pretty_profile(chi, lines, indent + 1)

    return lines


def _get_neo4j_version(driver) -> str | None:
    """
    Get Neo4j version from the database.

    Args:
        driver: Neo4j driver instance

    Returns:
        Version string or None if unavailable
    """
    if not NEO4J_AVAILABLE or not driver:
        return None

    try:
        with driver.session() as session:
            result = session.run(
                """
                CALL dbms.components()
                YIELD name, versions, edition
                UNWIND versions AS version
                RETURN version AS version
                """
            )
            data = result.data()
            if data:
                return data[0]["version"]
    except Exception as e:
        logger.warning(f"Error detecting Neo4j version: {e}")
    return None


class Neo4jDriver:
    """
    Manage the connection to the Neo4j server.

    A wrapper around the Neo4j driver that handles database connections
    and provides convenient methods for querying and managing the database.
    """

    _connect_essential = ("uri", "user", "passwd")

    def __init__(
        self,
        driver: neo4j.Driver | Neo4jDriver | None = None,
        db_name: str | None = None,
        db_uri: str | None = None,
        db_user: str | None = None,
        db_passwd: str | None = None,
        config: CONFIG_FILES | None = None,
        fetch_size: int = 1000,
        raise_errors: bool | None = None,
        wipe: bool = False,
        offline: bool = False,
        fallback_db: str | tuple[str] | None = None,
        fallback_on: str | set[str] | None = None,
        multi_db: bool | None = None,
        force_enterprise: bool = False,
        **kwargs,
    ):
        """
        Create a Driver object with database connection and runtime parameters.

        Args:
            driver:
                A neo4j.Driver instance, created by neo4j.GraphDatabase.driver.
            db_name:
                Name of the database (Neo4j graph) to use.
            db_uri:
                Protocol, host and port to access the Neo4j server.
            db_user:
                Neo4j user name.
            db_passwd:
                Password of the Neo4j user.
            fetch_size:
                Optional; the fetch size to use in database transactions.
            raise_errors:
                Raise the errors instead of turning them into log messages
                and returning None.
            config:
                Path to a YAML config file which provides the URI, user
                name and password.
            wipe:
                Wipe the database after connection, ensuring the data is
                loaded into an empty database.
            offline:
                Disable any interaction to the server. Queries won't be
                executed. The config will be still stored in the object
                and it will be ready to go online by its go_online method.
            fallback_db:
                Arbitrary number of fallback databases. If a query fails
                to run against the current database, it will be attempted
                against the fallback databases.
            fallback_on:
                Switch to the fallback databases upon these errors.
            multi_db:
                Whether to use multi-database mode (Neo4j 4.0+).
            kwargs:
                Ignored.
        """
        if not NEO4J_AVAILABLE:
            raise ImportError("Neo4j driver is not installed. Install it with: " "pip install neo4j>=5.0")

        self.driver = getattr(driver, "driver", driver)
        self._db_config = {
            "uri": db_uri,
            "user": db_user,
            "passwd": db_passwd,
            "db": db_name,
            "fetch_size": fetch_size,
            "raise_errors": raise_errors,
            "fallback_db": fallback_db,
            "fallback_on": fallback_on,
        }
        self._config_file = config
        self._drivers = {}
        self._queries = {}
        self._offline = offline
        self.multi_db = multi_db
        self._neo4j_version_cache = None
        self._force_enterprise = force_enterprise

        if self.driver:
            logger.info("Using the driver provided.")
            self._config_from_driver()
            self._register_current_driver()
        else:
            logger.info("No driver provided, initialising it from local config.")
            self.db_connect()

        # Detect Community Edition and adjust settings accordingly
        # Default to Community Edition (safer for CI) unless explicitly overridden
        self._detect_and_handle_community_edition()

        self.ensure_db()

        if wipe:
            self.wipe_db()

    def db_connect(self):
        """Connect to the database server."""
        if not self._connect_param_available:
            self.read_config()

        con_param = f"uri={self.uri}, auth=(user, ***)"
        logger.info(f"Attempting to connect: {con_param}")

        if self.offline:
            self.driver = None
            logger.info("Offline mode, not connecting to database.")
        else:
            self.driver = neo4j.GraphDatabase.driver(
                uri=self.uri,
                auth=self.auth,
            )
            logger.info("Opened database connection.")

        self._register_current_driver()

    def _detect_and_handle_community_edition(self):
        """
        Detect Community Edition and adjust settings for compatibility.

        Community Edition doesn't support multi-database, so we:
        1. Convert neo4j:// to bolt:// to avoid routing issues
        2. Disable multi_db mode
        3. Use default database 'neo4j' if a custom database was requested
        """
        if not self.driver or self.offline:
            return

        # If Enterprise Edition is forced, skip detection
        if self._force_enterprise:
            logger.info("Enterprise Edition mode forced. Skipping Community Edition detection.")
            return

        # Check if multi-database is supported (Enterprise Edition)
        # Use bolt:// for detection to avoid routing table issues
        original_uri = self.uri
        detection_uri = original_uri
        if original_uri.startswith("neo4j://"):
            detection_uri = original_uri.replace("neo4j://", "bolt://", 1)
        elif original_uri.startswith("neo4j+s://"):
            detection_uri = original_uri.replace("neo4j+s://", "bolt+s://", 1)

        # Create a temporary driver with bolt:// for detection
        temp_driver = None
        supports_multi_db = False
        try:
            temp_driver = neo4j.GraphDatabase.driver(uri=detection_uri, auth=self.auth)
            with temp_driver.session(database="neo4j") as session:
                result = session.run(
                    """
                    CALL dbms.components()
                    YIELD edition
                    RETURN edition CONTAINS 'enterprise' AS is_enterprise
                    """
                )
                data = result.data()
                supports_multi_db = data[0].get("is_enterprise", False) if data else False
        except Exception as e:
            logger.debug(f"Error detecting Neo4j edition: {e}. Assuming Community Edition.")
            # If detection fails, assume Community Edition (safer)
            supports_multi_db = False
        finally:
            if temp_driver:
                temp_driver.close()

        # If Community Edition or detection failed, adjust settings
        if not supports_multi_db:
            logger.info(
                "Neo4j Community Edition detected (or detection failed). "
                "Multi-database features are not available. "
                "Adjusting configuration for compatibility."
            )

            # Convert neo4j:// to bolt:// to avoid routing table issues
            # (already converted for detection, but need to update main driver)
            try:
                if original_uri.startswith("neo4j://"):
                    bolt_uri = original_uri.replace("neo4j://", "bolt://", 1)
                    self._db_config["uri"] = bolt_uri
                    logger.info(f"Converted URI from {original_uri} to {bolt_uri} for Community Edition compatibility.")
                    # Reconnect with bolt://
                    self.driver.close()
                    self.db_connect()
                elif original_uri.startswith("neo4j+s://"):
                    bolt_uri = original_uri.replace("neo4j+s://", "bolt+s://", 1)
                    self._db_config["uri"] = bolt_uri
                    logger.info(f"Converted URI from {original_uri} to {bolt_uri} for Community Edition compatibility.")
                    # Reconnect with bolt+s://
                    self.driver.close()
                    self.db_connect()
            except Exception as e:
                logger.warning(f"Failed to convert URI and reconnect: {e}. Continuing with original URI.")

            # Disable multi_db mode
            if self.multi_db:
                logger.info("Disabling multi-database mode for Community Edition.")
                self.multi_db = False

            # Use default database if a custom database was requested
            current_db = self.current_db
            if current_db and current_db.lower() != "neo4j":
                logger.warning(
                    f"Requested database '{current_db}' is not supported in Community Edition. "
                    f"Falling back to default database 'neo4j'."
                )
                self._db_config["db"] = "neo4j"
                self._register_current_driver()

    @property
    def _connect_param_available(self) -> bool:
        """Check for essential connection parameters."""
        return all(self._db_config.get(k, None) for k in self._connect_essential)

    @property
    def status(
        self,
    ) -> Literal[
        "no driver",
        "no connection",
        "db offline",
        "db online",
        "offline",
    ]:
        """State of this driver object and its current database."""
        if self.offline:
            return "offline"

        if not self.driver:
            return "no driver"

        db_status = self.db_status()
        return f"db {db_status}" if db_status else "no connection"

    @property
    def uri(self) -> str:
        """Database server URI (from config or built-in default)."""
        return self._db_config.get("uri") or DEFAULT_CONFIG["uri"]

    @property
    def auth(self) -> tuple[str, str]:
        """Database server user and password (from config or built-in default)."""
        auth_tuple = self._db_config.get("auth")
        if auth_tuple:
            return tuple(auth_tuple)
        return (
            self._db_config.get("user") or DEFAULT_CONFIG["user"],
            self._db_config.get("passwd") or DEFAULT_CONFIG["passwd"],
        )

    def read_config(self, section: str | None = None):
        """Read the configuration from a YAML file."""
        config_key_synonyms = {
            "password": "passwd",
            "pw": "passwd",
            "username": "user",
            "login": "user",
            "host": "uri",
            "address": "uri",
            "server": "uri",
            "graph": "db",
            "database": "db",
            "name": "db",
        }

        if not self._config_file or not os.path.exists(self._config_file):
            confdirs = (".", appdirs.user_config_dir("biocypher", "biocypher"))
            conffiles = ("neo4j.yaml", "neo4j.yml")

            for config_path_t in itertools.product(confdirs, conffiles):
                config_path_s = os.path.join(*config_path_t)
                if os.path.exists(config_path_s):
                    self._config_file = config_path_s

        if self._config_file and os.path.exists(self._config_file):
            logger.info(f"Reading config from `{self._config_file}`.")

            with open(self._config_file) as fp:
                conf = yaml.safe_load(fp.read())

            for k, v in conf.get(section, conf).items():
                k = k.lower()
                k = config_key_synonyms.get(k, k)

                if not self._db_config.get(k, None):
                    self._db_config[k] = v

        elif not self._connect_param_available:
            logger.warning("No config available, falling back to defaults.")

        self._config_from_defaults()

    def _config_from_driver(self):
        """Extract configuration from an existing driver."""
        from_driver = {
            "uri": self._uri(
                host=getattr(self.driver, "default_host", None),
                port=getattr(self.driver, "default_port", None),
            ),
            "db": self.current_db,
            "fetch_size": getattr(
                getattr(self.driver, "_default_workspace_config", None),
                "fetch_size",
                None,
            ),
            "user": self.user,
            "passwd": self.passwd,
        }

        for k, v in from_driver.items():
            self._db_config[k] = self._db_config.get(k, v) or v

        self._config_from_defaults()

    def _config_from_defaults(self):
        """Populate missing config items by their default values."""
        for k, v in DEFAULT_CONFIG.items():
            if self._db_config.get(k, None) is None:
                self._db_config[k] = v

    def _register_current_driver(self):
        """Register the current driver for the current database."""
        self._drivers[self.current_db] = self.driver

    @staticmethod
    def _uri(
        host: str = "localhost",
        port: str | int = 7687,
        protocol: str = "neo4j",
    ) -> str:
        """Construct a Neo4j URI."""
        return f"{protocol}://{host}:{port}/"

    def close(self):
        """Close the Neo4j driver if it exists and is open."""
        if hasattr(self, "driver") and hasattr(self.driver, "close"):
            self.driver.close()

    def __del__(self):
        """Cleanup on deletion."""
        self.close()

    @property
    def current_db(self) -> str:
        """Name of the current database."""
        return self._db_config["db"] or self._driver_con_db or self.home_db or neo4j.DEFAULT_DATABASE

    @current_db.setter
    def current_db(self, name: str):
        """Set the database currently in use."""
        self._db_config["db"] = name
        self.db_connect()

    @property
    def _driver_con_db(self) -> str | None:
        """Get the database from the driver connection."""
        if not self.driver:
            return None

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                driver_con = self.driver.verify_connectivity()
            except neo4j_exc.ServiceUnavailable:
                logger.error("Cannot access Neo4j server.")
                return None

        if driver_con:
            first_con = next(iter(driver_con.values()))[0]
            return first_con.get("db", None)

        return None

    @property
    def home_db(self) -> str | None:
        """Home database of the current user."""
        return self._db_name("HOME")

    @property
    def default_db(self) -> str | None:
        """Default database of the server."""
        return self._db_name("DEFAULT")

    def _db_name(self, which: Literal["HOME", "DEFAULT"] = "HOME") -> str | None:
        """Get the HOME or DEFAULT database name."""
        try:
            resp, summary = self.query(
                f"SHOW {which} DATABASE;",
                fallback_db=self._get_fallback_db,
            )
        except (neo4j_exc.AuthError, neo4j_exc.ServiceUnavailable) as e:
            logger.error(f"No connection to Neo4j server: {e}")
            return None

        if resp:
            return resp[0]["name"]
        return None

    @property
    def _get_fallback_db(self) -> tuple[str]:
        """Get fallback database tuple."""
        return _to_tuple(getattr(self, "_fallback_db", None) or self._db_config["fallback_db"])

    @property
    def _get_fallback_on(self) -> set[str]:
        """Get fallback error types."""
        return _to_set(getattr(self, "_fallback_on", None) or self._db_config["fallback_on"])

    def query(
        self,
        query: str,
        db: str | None = None,
        fetch_size: int | None = None,
        write: bool = True,
        explain: bool = False,
        profile: bool = False,
        fallback_db: str | tuple[str] | None = None,
        fallback_on: str | set[str] | None = None,
        raise_errors: bool | None = None,
        parameters: dict | None = None,
        **kwargs,
    ) -> tuple[list[dict] | None, neo4j.work.summary.ResultSummary | None]:
        """
        Run a Cypher query.

        Args:
            query:
                A valid Cypher query.
            db:
                The DB inside the Neo4j server that should be queried.
            fetch_size:
                The Neo4j fetch size parameter.
            write:
                Indicates whether to address write- or read-servers.
            explain:
                Indicates whether to EXPLAIN the Cypher query.
            profile:
                Indicates whether to PROFILE the Cypher query.
            fallback_db:
                If the query fails, try to execute it against a fallback database.
            fallback_on:
                Run queries against the fallback databases in case of these errors.
            raise_errors:
                Raise Neo4j errors instead of only printing them.
            parameters:
                Parameters dictionary for the query.
            **kwargs:
                Additional parameters (deprecated, use parameters dict instead).

        Returns:
            2-tuple:
                - neo4j.Record.data: the Neo4j response to the query
                - neo4j.ResultSummary: information about the result
        """
        if explain:
            query = "EXPLAIN " + query
        elif profile:
            query = "PROFILE " + query

        if self.offline:
            logger.info(f"Offline mode, not running query: `{query}`.")
            return None, None

        if not self.driver:
            if raise_errors:
                raise RuntimeError("Driver is not available. The driver may be closed or in offline mode.")
            logger.error("Driver is not available. Cannot execute query.")
            return None, None

        # Check if driver is closed (Neo4j 5.x driver has _closed attribute)
        if hasattr(self.driver, "_closed") and self.driver._closed:
            if raise_errors:
                raise RuntimeError("Driver is closed. Please reconnect or create a new driver instance.")
            logger.error("Driver is closed. Cannot execute query.")
            return None, None

        db = db or self._db_config["db"] or neo4j.DEFAULT_DATABASE
        fetch_size = fetch_size or self._db_config["fetch_size"]
        raise_errors = self._db_config["raise_errors"] if raise_errors is None else raise_errors

        # Combine parameters dict with kwargs (kwargs for backward compatibility)
        query_params = dict(parameters or {}, **kwargs)

        # Neo4j 5+ uses database parameter, older versions use it conditionally
        session_kwargs = {
            "fetch_size": fetch_size,
            "default_access_mode": (neo4j.WRITE_ACCESS if write else neo4j.READ_ACCESS),
        }

        # For Neo4j 4.0+, use database parameter if multi_db is True
        # For Neo4j 5.0+, always use database parameter
        if self.multi_db or self._is_neo4j_5_plus():
            session_kwargs["database"] = db

        try:
            with self.session(**session_kwargs) as session:
                # Neo4j driver expects parameters via the 'parameters' argument,
                # not unpacked as kwargs. This ensures query parameters are correctly
                # passed to the Cypher query and prevents conflicts with method parameters.
                if query_params:
                    res = session.run(query, parameters=query_params)
                else:
                    res = session.run(query)
                return res.data(), res.consume()

        except (neo4j_exc.Neo4jError, neo4j_exc.DriverError) as e:
            fallback_db = fallback_db or getattr(self, "_fallback_db", ())
            fallback_on = _to_set(_if_none(fallback_on, self._get_fallback_on))

            if self._match_error(e, fallback_on):
                for fdb in _to_tuple(fallback_db):
                    if fdb != db:
                        logger.warning(f"Running query against fallback database `{fdb}`.")
                        return self.query(
                            query=query,
                            db=fdb,
                            fetch_size=fetch_size,
                            write=write,
                            fallback_on=set(),
                            raise_errors=raise_errors,
                            parameters=query_params,
                        )

            logger.error(f"Failed to run query: {e.__class__.__name__}: {e}")
            logger.error(f"The error happened with this query: {query}")

            if e.__class__.__name__ == "AuthError":
                logger.error("Authentication error, switching to offline mode.")
                self.go_offline()

            if raise_errors:
                raise

            return None, None

    def _is_neo4j_5_plus(self) -> bool:
        """Check if Neo4j version is 5.0 or higher."""
        if self._neo4j_version_cache is None:
            version_str = _get_neo4j_version(self.driver)
            if version_str:
                try:
                    major_version = int(version_str.split(".")[0])
                    self._neo4j_version_cache = major_version >= 5
                except (ValueError, IndexError):
                    self._neo4j_version_cache = False
            else:
                self._neo4j_version_cache = False
        return self._neo4j_version_cache

    def explain(self, query, db=None, fetch_size=None, write=True, **kwargs):
        """
        Explain a query and pretty print the output.

        Args:
            query: Cypher query to explain
            db: Database name
            fetch_size: Fetch size
            write: Write access mode
            **kwargs: Query parameters

        Returns:
            2-tuple:
                - dict: the raw plan returned by the Neo4j bolt driver
                - list of str: a list of strings ready for printing
        """
        logger.info("Explaining a query.")
        data, summary = self.query(query, db, fetch_size, write, explain=True, **kwargs)

        if not summary:
            return None, []

        plan = summary.plan
        printout = _pretty_profile(plan)

        return plan, printout

    def profile(self, query, db=None, fetch_size=None, write=True, **kwargs):
        """
        Profile a query and pretty print the output.

        Args:
            query: Cypher query to profile
            db: Database name
            fetch_size: Fetch size
            write: Write access mode
            **kwargs: Query parameters

        Returns:
            2-tuple:
                - dict: the raw profile returned by the Neo4j bolt driver
                - list of str: a list of strings ready for printing
        """
        logger.info("Profiling a query.")
        data, summary = self.query(query, db, fetch_size, write, profile=True, **kwargs)

        if not summary:
            return None, []

        prof = summary.profile
        exec_time = summary.result_available_after + summary.result_consumed_after

        # get print representation
        header = f"Execution time: {exec_time:n}\n"
        printout = _pretty_profile(prof, [header], indent=0)

        return prof, printout

    def db_exists(self, name: str | None = None) -> bool:
        """Check if a database exists."""
        return bool(self.db_status(name=name))

    def db_status(
        self,
        name: str | None = None,
        field: str = "currentStatus",
    ) -> Literal["online", "offline"] | str | dict | None:
        """
        Get the current status or other state info of a database.

        Args:
            name: Name of a database
            field: The field to return

        Returns:
            The status as a string, None if the database does not exist.
            If field is None, a dictionary with all fields will be returned.
        """
        name = name or self.current_db
        query = f'SHOW DATABASES WHERE name = "{name}";'

        # Use fallback context manager like original neo4j_utils
        # This allows query to default to current_db and fallback to system/neo4j on error
        with self.fallback():
            resp, summary = self.query(query)

        if resp:
            return resp[0].get(field, resp[0])
        return None

    def db_online(self, name: str | None = None) -> bool:
        """Check if a database is currently online."""
        return self.db_status(name=name) == "online"

    def create_db(self, name: str | None = None):
        """Create a database if it does not already exist."""
        self._manage_db("CREATE", name=name, options="IF NOT EXISTS")

    def start_db(self, name: str | None = None):
        """Start a database (bring it online) if it is offline."""
        self._manage_db("START", name=name)

    def stop_db(self, name: str | None = None):
        """Stop a database, making sure it's offline."""
        self._manage_db("STOP", name=name)

    def drop_db(self, name: str | None = None):
        """Delete a database if it exists."""
        self._manage_db("DROP", name=name, options="IF EXISTS")

    def _manage_db(
        self,
        cmd: Literal["CREATE", "START", "STOP", "DROP"],
        name: str | None = None,
        options: str | None = None,
    ):
        """Execute a database management command."""
        # Use fallback_db like original neo4j_utils
        # Query defaults to current_db, but fallback mechanism will retry against system/neo4j
        self.query(
            f"{cmd} DATABASE {name or self.current_db} {options or ''};",
            fallback_db=self._get_fallback_db,
        )

    def wipe_db(self):
        """Delete all contents of the current database."""
        if not self.driver:
            raise RuntimeError(
                "Driver is not available. Cannot wipe database. " "The driver may be closed or in offline mode."
            )

        # Check if driver is closed (Neo4j 5.x driver has _closed attribute)
        if hasattr(self.driver, "_closed") and self.driver._closed:
            raise RuntimeError(
                "Driver is closed. Cannot wipe database. " "Please reconnect or create a new driver instance."
            )

        # Ensure database exists before trying to wipe it
        self.ensure_db()

        # For Community Edition, use default database if current_db is not supported
        # Skip this check if Enterprise Edition is forced
        db_to_wipe = self.current_db
        if not self._force_enterprise:
            current_uri = self.uri
            is_neo4j_protocol = current_uri.startswith("neo4j://") or current_uri.startswith("neo4j+s://")
            is_non_default_db = db_to_wipe and db_to_wipe.lower() != "neo4j"
            is_community_edition = not self.multi_db or (is_neo4j_protocol and is_non_default_db)

            if is_community_edition and is_non_default_db:
                logger.warning(
                    f"Cannot wipe database '{db_to_wipe}' in Community Edition. "
                    f"Using default database 'neo4j' instead. "
                    f"Database will remain 'neo4j' for this session."
                )
                # Permanently change to default database for Community Edition
                db_to_wipe = "neo4j"
                self._db_config["db"] = "neo4j"
                self._register_current_driver()

        logger.info(f"Wiping database `{db_to_wipe}`.")
        self.query("MATCH (n) DETACH DELETE n;")
        self.drop_indices_constraints()

    def ensure_db(self):
        """Make sure the database exists and is online."""
        db_name = self.current_db

        # Skip if offline mode
        if self.offline:
            logger.debug(f"Offline mode, skipping database creation for '{db_name}'.")
            return

        # If Enterprise Edition is forced, skip Community Edition checks
        if self._force_enterprise:
            logger.debug(f"Enterprise Edition forced, proceeding with database check for '{db_name}'.")
            # Continue to database existence check below
        else:
            # In Community Edition, multi-database operations are not supported
            # The default database 'neo4j' always exists and is always online
            # Also skip if URI is bolt:// (which indicates Community Edition or direct connection)
            # If URI is still neo4j:// and we're checking a non-default database, assume Community Edition
            current_uri = self.uri
            is_bolt = current_uri.startswith("bolt://") or current_uri.startswith("bolt+s://")
            is_neo4j_protocol = current_uri.startswith("neo4j://") or current_uri.startswith("neo4j+s://")
            is_non_default_db = db_name and db_name.lower() != "neo4j"

            if not self.multi_db or is_bolt or (is_neo4j_protocol and is_non_default_db):
                if not self.multi_db:
                    logger.debug(
                        f"Multi-database mode disabled (Community Edition). "
                        f"Using default database '{db_name}' which always exists."
                    )
                elif is_bolt:
                    logger.debug(
                        f"Using bolt:// connection (direct mode). "
                        f"Using default database '{db_name}' which always exists."
                    )
                else:
                    logger.debug(
                        f"Using neo4j:// protocol with non-default database '{db_name}' - "
                        f"assuming Community Edition and skipping database check."
                    )
                return

        # Check if database exists, create if needed
        try:
            exists = self.db_exists()
            if not exists:
                logger.info(f"Database '{db_name}' does not exist, creating it...")
                self.create_db()
                # Verify creation succeeded
                if not self.db_exists():
                    raise RuntimeError(
                        f"Failed to create database '{db_name}'. " "The database was not created successfully."
                    )
                logger.info(f"Database '{db_name}' created successfully.")
            else:
                logger.debug(f"Database '{db_name}' already exists.")
        except Exception as e:
            logger.error(f"Failed to check/create database '{db_name}': {e}")
            # Re-raise to prevent initialization from continuing with a missing database
            raise RuntimeError(
                f"Failed to ensure database '{db_name}' exists: {e}. "
                "Please check Neo4j permissions and that the database can be created."
            ) from e

        # Check if database is online, start if needed
        try:
            if not self.db_online():
                logger.info(f"Database '{db_name}' is offline, starting it...")
                self.start_db()
                # Verify start succeeded
                if not self.db_online():
                    raise RuntimeError(
                        f"Failed to start database '{db_name}'. " "The database was not started successfully."
                    )
                logger.info(f"Database '{db_name}' started successfully.")
            else:
                logger.debug(f"Database '{db_name}' is already online.")
        except Exception as e:
            logger.error(f"Failed to check/start database '{db_name}': {e}")
            # Re-raise to prevent initialization from continuing with an offline database
            raise RuntimeError(
                f"Failed to ensure database '{db_name}' is online: {e}. "
                "Please check Neo4j permissions and that the database can be started."
            ) from e

    def select_db(self, name: str):
        """Set the current database."""
        current = self.current_db

        if current != name:
            self._register_current_driver()
            self._db_config["db"] = name

            if name in self._drivers:
                self.driver = self._drivers[name]
            else:
                self.db_connect()

    @property
    def indices(self) -> list | None:
        """List of indices in the current database."""
        return self._list_indices("indices")

    @property
    def constraints(self) -> list | None:
        """List of constraints in the current database."""
        return self._list_indices("constraints")

    def drop_indices_constraints(self):
        """Drop all indices and constraints in the current database."""
        # Neo4j 5+ handles constraints and indexes together
        self.drop_constraints()
        # For older versions, also drop indexes separately
        if not self._is_neo4j_5_plus():
            self.drop_indices()

    def drop_constraints(self):
        """Drop all constraints in the current database."""
        self._drop_indices(what="constraints")

    def drop_indices(self):
        """Drop all indices in the current database."""
        self._drop_indices(what="indexes")

    def _drop_indices(
        self,
        what: Literal["indexes", "indices", "constraints"] = "constraints",
    ):
        """Drop indices or constraints.

        Compatible with Neo4j 4.x and 5.x. Uses SHOW syntax which is
        available in both versions.
        """
        what_u = self._idx_cstr_synonyms(what)

        with self.session() as s:
            try:
                # SHOW INDEXES and SHOW CONSTRAINTS work in both Neo4j 4.x and 5.x
                # Neo4j 5.x unified constraints and indexes, but separate commands still work
                if what == "constraints":
                    query = "SHOW CONSTRAINTS"
                elif what in ("indexes", "indices"):
                    query = "SHOW INDEXES"
                else:
                    query = f"SHOW {what_u}S"  # Plural form

                indices = s.run(query)
                indices = list(indices)
                n_indices = len(indices)
                index_names = ", ".join(i["name"] for i in indices)

                for idx in indices:
                    s.run(f"DROP {what_u} `{idx['name']}` IF EXISTS")

                logger.info(f"Dropped {n_indices} {what}: {index_names}.")

            except (neo4j_exc.Neo4jError, neo4j_exc.DriverError) as e:
                logger.error(f"Failed to run query: {e}")

    def _list_indices(
        self,
        what: Literal["indexes", "indices", "constraints"] = "constraints",
    ) -> list | None:
        """List indices or constraints."""
        what_u = self._idx_cstr_synonyms(what)

        with self.session() as s:
            try:
                return list(s.run(f"SHOW {what_u.upper()};"))
            except (neo4j_exc.Neo4jError, neo4j_exc.DriverError) as e:
                logger.error(f"Failed to run query: {e}")
                return None

    @staticmethod
    def _idx_cstr_synonyms(what: str) -> str:
        """Convert index/constraint keyword to Cypher keyword."""
        what_s = {
            "indexes": "INDEX",
            "indices": "INDEX",
            "constraints": "CONSTRAINT",
        }

        what_u = what_s.get(what, None)

        if not what_u:
            msg = f'Allowed keywords are: "indexes", "indices" or "constraints", ' f"not `{what}`."
            logger.error(msg)
            raise ValueError(msg)

        return what_u

    @property
    def node_count(self) -> int | None:
        """Number of nodes in the database."""
        res, summary = self.query("MATCH (n) RETURN COUNT(n) AS count;")
        return res[0]["count"] if res else None

    @property
    def edge_count(self) -> int | None:
        """Number of edges in the database."""
        res, summary = self.query("MATCH ()-[r]->() RETURN COUNT(r) AS count;")
        return res[0]["count"] if res else None

    @property
    def user(self) -> str | None:
        """User for the currently active connection."""
        return self._extract_auth[0]

    @property
    def passwd(self) -> str | None:
        """Password for the currently active connection."""
        return self._extract_auth[1]

    @property
    def _extract_auth(self) -> tuple[str | None, str | None]:
        """Extract authentication data from the Neo4j driver."""
        auth = None, None

        if self.driver:
            opener_vars = self._opener_vars
            if "auth" in opener_vars:
                auth = opener_vars["auth"].cell_contents

        return auth

    @property
    def _opener_vars(self) -> dict:
        """Extract variables from the opener part of the Neo4j driver."""
        return dict(
            zip(
                self.driver._pool.opener.__code__.co_freevars,
                self.driver._pool.opener.__closure__,
            ),
        )

    def __len__(self):
        """Return the number of nodes in the database."""
        return self.node_count or 0

    @contextlib.contextmanager
    def use_db(self, name: str):
        """Context manager where the default database is set to name."""
        used_previously = self.current_db
        self.select_db(name=name)

        try:
            yield None
        finally:
            self.select_db(name=used_previously)

    @contextlib.contextmanager
    def fallback(
        self,
        db: str | tuple[str] | None = None,
        on: str | set[str] | None = None,
    ):
        """
        Context manager that attempts to run queries against a fallback database
        if running against the default database fails.
        """
        prev = {}

        for var in ("db", "on"):
            prev[var] = getattr(self, f"_fallback_{var}", None)
            setattr(
                self,
                f"_fallback_{var}",
                locals()[var] or self._db_config.get(f"fallback_{var}"),
            )

        try:
            yield None
        finally:
            for var in ("db", "on"):
                setattr(self, f"_fallback_{var}", prev[var])

    @contextlib.contextmanager
    def session(self, **kwargs):
        """Context manager with a database connection session."""
        if not self.driver:
            raise RuntimeError("Driver is not available. The driver may be closed or in offline mode.")

        # Check if driver is closed
        if hasattr(self.driver, "_closed") and self.driver._closed:
            raise RuntimeError("Driver is closed. Please reconnect or create a new driver instance.")

        session = self.driver.session(**kwargs)

        try:
            yield session
        finally:
            session.close()

    def __enter__(self):
        """Context manager entry."""
        self._context_session = self.session()
        return self._context_session

    def __exit__(self, *exc):
        """Context manager exit."""
        if hasattr(self, "_context_session"):
            self._context_session.close()
            delattr(self, "_context_session")

    def __repr__(self):
        """String representation."""
        return f"<{self.__class__.__name__} " f"{self._connection_str if self.driver else '[offline]'}>"

    @property
    def _connection_str(self) -> str:
        """Connection string representation."""
        if not self.driver:
            return "unknown://unknown:0/unknown"

        protocol = re.split(
            r"(?<=[a-z])(?=[A-Z])",
            self.driver.__class__.__name__,
        )[0].lower()

        address = self.driver._pool.address if hasattr(self.driver, "_pool") else ("unknown", 0)

        return f"{protocol}://{address[0]}:{address[1]}/{self.user or 'unknown'}"

    @property
    def offline(self) -> bool:
        """Whether the driver is in offline mode."""
        return self._offline

    @offline.setter
    def offline(self, offline: bool):
        """Enable or disable offline mode."""
        self.go_offline() if offline else self.go_online()

    @property
    def apoc_version(self) -> str | None:
        """
        Version of the APOC plugin available in the current database.

        Returns:
            APOC version string or None if APOC is not available
        """
        # Check if driver is available before attempting to query
        if not self.driver or self.offline:
            return None

        # Check if driver is closed
        if hasattr(self.driver, "_closed") and self.driver._closed:
            return None

        db = self._db_config["db"] or neo4j.DEFAULT_DATABASE

        try:
            with self.session(database=db) as session:
                res = session.run("RETURN apoc.version() AS output;")
                data = res.data()
                if data:
                    return data[0]["output"]
        except (neo4j_exc.ClientError, RuntimeError):
            # RuntimeError can be raised if driver is offline/closed
            # ClientError is raised if APOC is not available
            return None
        except Exception:
            # Catch any other exceptions (e.g., connection errors) and return None
            return None
        return None

    @property
    def has_apoc(self) -> bool:
        """
        Check if APOC is available in the current database.

        Returns:
            True if APOC is available, False otherwise
        """
        try:
            return bool(self.apoc_version)
        except Exception:
            # Ensure has_apoc always returns a boolean, even if apoc_version raises
            return False

    def go_offline(self):
        """Switch to offline mode."""
        self._offline = True
        self.close()
        self.driver = None
        self._register_current_driver()
        logger.warning("Offline mode: any interaction to the server is disabled.")

    def go_online(
        self,
        db_name: str | None = None,
        db_uri: str | None = None,
        db_user: str | None = None,
        db_passwd: str | None = None,
        config: CONFIG_FILES | None = None,
        fetch_size: int | None = None,
        raise_errors: bool | None = None,
        wipe: bool = False,
    ):
        """Switch to online mode."""
        self._offline = False

        try:
            for k, current in self._db_config.items():
                self._db_config[k] = _if_none(
                    locals().get(k.replace("db_", ""), None),
                    current,
                    DEFAULT_CONFIG.get(k),
                )

            self._config_file = self._config_file or config

            self.db_connect()
            self.ensure_db()
            logger.info("Online mode: ready to run queries.")

        except Exception as e:
            logger.error(f"Failed to connect: {e}")
            self._offline = True

        if wipe:
            self.wipe_db()

    @staticmethod
    def _match_error(error: Exception | str, errors: set[Exception | str]) -> bool:
        """Check if error is listed in errors."""
        import builtins

        def str_to_exc(e):
            if isinstance(e, Exception):
                return e.__class__
            elif isinstance(e, str):
                return getattr(builtins, e, getattr(neo4j_exc, e, e))
            else:
                return e

        error = str_to_exc(error)
        errors = {str_to_exc(e) for e in _to_set(errors)}

        return error in errors or (
            isinstance(error, type) and any(issubclass(error, e) for e in errors if isinstance(e, type))
        )



import neo4j
import pytest

from biocypher._create import BioCypherEdge, BioCypherNode, BioCypherRelAsNode
from biocypher.output.connect._neo4j_driver import _Neo4jDriver


@pytest.mark.requires_neo4j
def test_create_driver(driver):
    assert isinstance(driver, _Neo4jDriver)


@pytest.mark.requires_neo4j
def test_connect_to_db(driver):
    assert isinstance(driver._driver.driver, neo4j.Neo4jDriver)


@pytest.mark.requires_neo4j
def test_increment_version(driver):
    driver._driver.wipe_db()
    query = "CREATE (n:BioCypher {id: 'v19700101-000000'})"
    driver._driver.query(query)
    driver._update_meta_graph()

    result, summary = driver._driver.query("MATCH (n:BioCypher) RETURN n")

    assert len(result) == 2


@pytest.mark.requires_neo4j
def test_explain(driver):
    query = "MATCH (n) WITH n LIMIT 25 MATCH (n)--(m)--(f) RETURN n, m, f"
    explanation = driver._driver.explain(query)
    text = explanation[0]

    assert "args" in text and "identifiers" in text


@pytest.mark.requires_neo4j
def test_profile(driver):
    query = "MATCH (n) RETURN n LIMIT 100"
    profile = driver._driver.profile(query)
    text = profile[0]

    assert "args" in text and "identifiers" in text


@pytest.mark.requires_neo4j
def test_add_invalid_biocypher_node(driver):
    # neo4j database needs to be running!

    with pytest.raises(ValueError):
        driver.add_biocypher_nodes(1)

    with pytest.raises(ValueError):
        driver.add_biocypher_nodes("String")


@pytest.mark.requires_neo4j
def test_add_single_biocypher_node(driver):
    # neo4j database needs to be running!
    node = BioCypherNode(node_id="test_id1", node_label="Test")
    driver.add_biocypher_nodes(node)
    result, summary = driver._driver.query(
        "MATCH (n:Test) WITH n, n.id AS id RETURN id ",
    )
    assert result[0]["id"] == "test_id1"


@pytest.mark.requires_neo4j
def test_add_biocypher_node_list(driver):
    # neo4j database needs to be running!
    node_1 = BioCypherNode(node_id="test_id1", node_label="Test")
    node_2 = BioCypherNode(node_id="test_id2", node_label="Test")
    driver.add_biocypher_nodes([node_1, node_2])
    result, summary = driver._driver.query(
        "MATCH (n:Test) WITH n, n.id AS id RETURN id ",
    )
    assert set([result[0]["id"], result[1]["id"]]) == set(["test_id1", "test_id2"])


@pytest.mark.requires_neo4j
def test_add_biocypher_node_generator(driver):
    # neo4j database needs to be running!
    # generator
    def gen(nodes):
        for g in nodes:
            yield BioCypherNode(g[0], g[1])

    node_generator = gen([("test_id1", "Test"), ("test_id2", "Test")])

    driver.add_biocypher_nodes(node_generator)
    result, summary = driver._driver.query(
        "MATCH (n:Test) WITH n, n.id AS id RETURN id ",
    )

    ids = [n["id"] for n in result]

    assert "test_id1" in ids
    assert "test_id2" in ids


@pytest.mark.requires_neo4j
def test_add_specific_id_node(driver):
    node = BioCypherNode(node_id="CHAT", node_label="Gene", preferred_id="hgnc")
    driver.add_biocypher_nodes(node)

    result, summary = driver._driver.query(
        "MATCH (n:Gene) RETURN n",
    )

    assert result[0]["n"].get("id") == "CHAT"
    assert result[0]["n"].get("preferred_id") == "hgnc"


@pytest.mark.requires_neo4j
def test_add_generic_id_node(driver):
    node = BioCypherNode(node_id="CHAT", node_label="Gene", preferred_id="HGNC")
    driver.add_biocypher_nodes(node)

    result, summary = driver._driver.query(
        "MATCH (n:Gene) RETURN n",
    )

    assert result[0]["n"].get("id") is not None


@pytest.mark.requires_neo4j
def test_add_invalid_biocypher_edge(driver):
    # neo4j database needs to be running!
    with pytest.raises(ValueError):
        driver.add_biocypher_edges([1, 2, 3])


@pytest.mark.requires_neo4j
def test_add_single_biocypher_edge_explicit_node_creation(driver):
    # neo4j database needs to be running!
    node_1 = BioCypherNode("src", "Test")
    node_2 = BioCypherNode("tar", "Test")
    driver.add_biocypher_nodes([node_1, node_2])

    edge = BioCypherEdge("src", "tar", "Test")
    driver.add_biocypher_edges(edge)
    result, summary = driver._driver.query(
        "MATCH (n1)-[r:Test]->(n2) "
        "WITH n1, n2, n1.id AS id1, n2.id AS id2, type(r) AS label "
        "RETURN id1, id2, label",
    )
    assert result[0]["id1"] == "src" and result[0]["id2"] == "tar" and result[0]["label"] == "Test"


@pytest.mark.requires_neo4j
def test_add_single_biocypher_edge_missing_nodes(driver):
    # neo4j database needs to be running!
    # merging on non-existing nodes creates them without labels; what is
    # the desired behaviour here? do we only want to MATCH?

    edge = BioCypherEdge("src", "tar", "Test")
    driver.add_biocypher_edges(edge)
    result, summary = driver._driver.query(
        "MATCH (n1)-[r:Test]->(n2) "
        "WITH n1, n2, n1.id AS id1, n2.id AS id2, type(r) AS label "
        "RETURN id1, id2, label",
    )
    assert result[0]["id1"] == "src" and result[0]["id2"] == "tar" and result[0]["label"] == "Test"


@pytest.mark.requires_neo4j
def test_add_biocypher_edge_list(driver):
    # neo4j database needs to be running!
    node_1 = BioCypherNode("src", "Test")
    node_2 = BioCypherNode("tar1", "Test")
    node_3 = BioCypherNode("tar2", "Test")
    driver.add_biocypher_nodes([node_1, node_2, node_3])

    # edge list
    edge_1 = BioCypherEdge("src", "tar1", "Test1")
    edge_2 = BioCypherEdge("src", "tar2", "Test2")
    driver.add_biocypher_edges([edge_1, edge_2])
    result, summary = driver._driver.query(
        "MATCH (n3)<-[r2:Test2]-(n1)-[r1:Test1]->(n2) "
        "WITH n1, n2, n3, n1.id AS id1, n2.id AS id2, n3.id AS id3, "
        "type(r1) AS label1, type(r2) AS label2 "
        "RETURN id1, id2, id3, label1, label2",
    )
    assert (
        result[0]["id1"] == "src"
        and result[0]["id2"] == "tar1"
        and result[0]["id3"] == "tar2"
        and result[0]["label1"] == "Test1"
        and result[0]["label2"] == "Test2"
    )


@pytest.mark.requires_neo4j
def test_add_biocypher_edge_generator(driver):
    # neo4j database needs to be running!
    node_1 = BioCypherNode("src", "Test")
    node_2 = BioCypherNode("tar1", "Test")
    node_3 = BioCypherNode("tar2", "Test")
    driver.add_biocypher_nodes([node_1, node_2, node_3])

    # generator
    def gen(edges):
        for e in edges:
            yield BioCypherEdge(
                e.get_source_id(),
                e.get_target_id(),
                e.get_label(),
            )

    # edge list
    edge_1 = BioCypherEdge("src", "tar1", "Test1")
    edge_2 = BioCypherEdge("src", "tar2", "Test2")
    edge_generator = gen([edge_1, edge_2])

    driver.add_biocypher_edges(edge_generator)
    result, summary = driver._driver.query(
        "MATCH (n3)<-[r2:Test2]-(n1)-[r1:Test1]->(n2) "
        "WITH n1, n2, n3, n1.id AS id1, n2.id AS id2, n3.id AS id3, "
        "type(r1) AS label1, type(r2) AS label2 "
        "RETURN id1, id2, id3, label1, label2",
    )
    assert (
        result[0]["id1"] == "src"
        and result[0]["id2"] == "tar1"
        and result[0]["id3"] == "tar2"
        and result[0]["label1"] == "Test1"
        and result[0]["label2"] == "Test2"
    )


@pytest.mark.requires_neo4j
def test_add_biocypher_interaction_as_BioCypherRelAsNode_list(driver):
    # neo4j database needs to be running!
    interaction_node_1 = BioCypherNode("int1", "Int1")
    interaction_node_2 = BioCypherNode("int2", "Int2")
    driver.add_biocypher_nodes([interaction_node_1, interaction_node_2])
    edge_1 = BioCypherEdge("src", "int1", "is_source_of")
    edge_2 = BioCypherEdge("tar", "int1", "is_target_of")
    edge_3 = BioCypherEdge("src", "int2", "is_source_of")
    edge_4 = BioCypherEdge("tar", "int2", "is_target_of")
    relationship_1, relationship_2 = (
        BioCypherRelAsNode(interaction_node_1, edge_1, edge_2),
        BioCypherRelAsNode(interaction_node_2, edge_3, edge_4),
    )
    driver.add_biocypher_edges([relationship_1, relationship_2])
    result, summary = driver._driver.query(
        "MATCH (n2)-[e4:is_target_of]->(i2:Int2)<-[e3:is_source_of]-"
        "(n1)-[e1:is_source_of]->(i1:Int1)<-[e2:is_target_of]-(n2)"
        "WITH n1, n2, i1, i2, n1.id AS id1, n2.id AS id2, "
        "i1.id AS id3, i2.id AS id4, "
        "type(e1) AS label1, type(e2) AS label2, "
        "type(e3) AS label3, type(e4) AS label4 "
        "RETURN id1, id2, id3, id4, label1, label2, label3, label4",
    )
    assert (
        result[0]["id1"] == "src"
        and result[0]["id2"] == "tar"
        and result[0]["id3"] == "int1"
        and result[0]["id4"] == "int2"
        and result[0]["label1"] == "is_source_of"
        and result[0]["label2"] == "is_target_of"
        and result[0]["label3"] == "is_source_of"
        and result[0]["label4"] == "is_target_of"
    )


@pytest.mark.requires_neo4j
def test_add_biocypher_interaction_as_BioCypherRelAsNode_generator(driver):
    # neo4j database needs to be running!
    interaction_node_1 = BioCypherNode("int1", "Int1")
    interaction_node_2 = BioCypherNode("int2", "Int2")
    driver.add_biocypher_nodes([interaction_node_1, interaction_node_2])
    edge_1 = BioCypherEdge("src", "int1", "is_source_of")
    edge_2 = BioCypherEdge("tar", "int1", "is_target_of")
    edge_3 = BioCypherEdge("src", "int2", "is_source_of")
    edge_4 = BioCypherEdge("tar", "int2", "is_target_of")
    relationship_1, relationship_2 = (
        BioCypherRelAsNode(interaction_node_1, edge_1, edge_2),
        BioCypherRelAsNode(interaction_node_2, edge_3, edge_4),
    )
    relasnode_list = [relationship_1, relationship_2]

    def gen(lis):
        for tup in lis:
            yield tup

    driver.add_biocypher_edges(gen(relasnode_list))
    result, summary = driver._driver.query(
        "MATCH (n2)-[e4:is_target_of]->(i2:Int2)<-[e3:is_source_of]-"
        "(n1)-[e1:is_source_of]->(i1:Int1)<-[e2:is_target_of]-(n2)"
        "WITH n1, n2, i1, i2, n1.id AS id1, n2.id AS id2, "
        "i1.id AS id3, i2.id AS id4, "
        "type(e1) AS label1, type(e2) AS label2, "
        "type(e3) AS label3, type(e4) AS label4 "
        "RETURN id1, id2, id3, id4, label1, label2, label3, label4",
    )
    assert (
        result[0]["id1"] == "src"
        and result[0]["id2"] == "tar"
        and result[0]["id3"] == "int1"
        and result[0]["id4"] == "int2"
        and result[0]["label1"] == "is_source_of"
        and result[0]["label2"] == "is_target_of"
        and result[0]["label3"] == "is_source_of"
        and result[0]["label4"] == "is_target_of"
    )


@pytest.mark.requires_neo4j
def test_pretty_profile(driver):
    prof, printout = driver._driver.profile(
        "UNWIND [1,2,3,4,5] as id MERGE (n:Test {id: id}) MERGE (x:Test {id: id + 1})",
    )

    assert "args" in prof and "ProduceResults" in printout[1]


@pytest.mark.requires_neo4j
def test_pretty_explain(driver):
    plan, printout = driver._driver.explain(
        "UNWIND [1,2,3,4,5] as id MERGE (n:Test {id: id}) MERGE (x:Test {id: id + 1})",
    )

    assert "args" in plan and "ProduceResults" in printout[0]



import pandas as pd

from biocypher.output.in_memory._in_memory_kg import _InMemoryKG


class PandasKG(_InMemoryKG):
    def __init__(self, deduplicator):
        super().__init__()  # keeping in spite of ABC not having __init__
        self.deduplicator = deduplicator

        self.dfs = {}

    def get_kg(self):
        return self.dfs

    def add_nodes(self, nodes):
        self.add_tables(nodes)

    def add_edges(self, edges):
        self.add_tables(edges)

    def add_tables(self, entities):
        """Add Pandas dataframes for each node and edge type in the input."""
        lists = self._separate_entity_types(entities)

        for _type, _entities in lists.items():
            self._add_entity_df(_type, _entities)

    def _add_entity_df(self, _type, _entities):
        df = pd.DataFrame(pd.json_normalize([node.get_dict() for node in _entities]))
        # replace "properties." with "" in column names
        df.columns = [col.replace("properties.", "") for col in df.columns]
        if _type not in self.dfs:
            self.dfs[_type] = df
        else:
            self.dfs[_type] = pd.concat([self.dfs[_type], df], ignore_index=True)
        return self.dfs[_type]



import json
import os

from datetime import datetime, timedelta
from unittest.mock import patch, Mock

import pytest

from hypothesis import (
    given,
    strategies as st,
)

from biocypher._get import APIRequest, Downloader, FileDownload, Resource


@pytest.fixture
def downloader(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def downloader_without_specified_cache_dir():
    return Downloader()


@pytest.fixture
def downloader_with_specified_cache_dir(tmp_path):
    tmp_cache_dir = tmp_path / ".cache"
    tmp_cache_dir.mkdir()
    return Downloader(cache_dir=str(tmp_cache_dir))


@given(
    st.builds(
        Resource,
        name=st.text(),
        url_s=st.text(),
        lifetime=st.integers(),
    )
)
def test_resource(resource):
    assert isinstance(resource.name, str)
    assert isinstance(resource.url_s, str) or isinstance(resource.url_s, list)
    assert isinstance(resource.lifetime, int)


@given(
    st.builds(
        FileDownload,
        name=st.text(),
        url_s=st.text(),
        lifetime=st.integers(),
    )
)
def test_file_download(file_download):
    assert isinstance(file_download.name, str)
    assert isinstance(file_download.url_s, str)
    assert isinstance(file_download.lifetime, int)


@given(
    st.builds(
        APIRequest,
        name=st.text(),
        url_s=st.text(),
        lifetime=st.integers(),
    )
)
def test_API(api):
    assert isinstance(api.name, str)
    assert isinstance(api.url_s, str)
    assert isinstance(api.lifetime, int)


@pytest.mark.parametrize(
    "downloader",
    [
        "downloader_without_specified_cache_dir",
        "downloader_with_specified_cache_dir",
    ],
    indirect=True,
)
def test_downloader(downloader):
    assert isinstance(downloader.cache_dir, str)
    assert isinstance(downloader.cache_file, str)


@pytest.mark.parametrize(
    "downloader",
    [
        "downloader_without_specified_cache_dir",
        "downloader_with_specified_cache_dir",
    ],
    indirect=True,
)
def test_download_file(downloader):
    resource = FileDownload(
        "test_resource",
        "https://github.com/biocypher/biocypher/raw/main/biocypher/_config/test_config.yaml",
        lifetime=7,
    )
    paths = downloader.download(resource)
    initial_download_time = os.path.getmtime(paths[0])
    assert len(paths) == 1
    assert os.path.exists(paths[0])
    assert f"{os.sep}test_resource{os.sep}test_config.yaml" in paths[0]

    # test caching
    paths = downloader.download(resource)
    assert len(paths) == 1
    # should not download again
    assert initial_download_time == os.path.getmtime(paths[0])

    # manipulate cache dict to test expiration
    downloader.cache_dict["test_resource"]["date_downloaded"] = str(datetime.now() - timedelta(days=8))

    paths = downloader.download(resource)
    assert len(paths) == 1
    # should download again
    assert initial_download_time < os.path.getmtime(paths[0])


@pytest.mark.parametrize(
    "downloader",
    [
        "downloader_without_specified_cache_dir",
        "downloader_with_specified_cache_dir",
    ],
    indirect=True,
)
def test_download_lists(downloader):
    resource1 = FileDownload(
        name="test_resource1",
        url_s=[
            "https://github.com/biocypher/biocypher/raw/main/biocypher/_config/test_config.yaml",
            "https://github.com/biocypher/biocypher/raw/main/biocypher/_config/test_schema_config_disconnected.yaml",
        ],
    )
    resource2 = FileDownload(
        "test_resource2",
        "https://github.com/biocypher/biocypher/raw/main/test/test_CSVs.zip",
    )
    paths = downloader.download(resource1, resource2)
    assert len(paths) == 4  # 2 files from resource1, 2 files from resource2 zip
    assert os.path.exists(paths[0])
    assert os.path.exists(paths[1])
    assert os.path.exists(paths[2])
    assert os.path.exists(paths[3])
    expected_paths = [
        os.path.realpath(os.path.join(downloader.cache_dir, "test_resource1", "test_config.yaml")),
        os.path.realpath(
            os.path.join(
                downloader.cache_dir,
                "test_resource1",
                "test_schema_config_disconnected.yaml",
            )
        ),
        os.path.realpath(
            os.path.join(
                downloader.cache_dir,
                "test_resource2",
                "test_CSVs.zip.unzip",
                "file1.csv",
            )
        ),
        os.path.realpath(
            os.path.join(
                downloader.cache_dir,
                "test_resource2",
                "test_CSVs.zip.unzip",
                "file2.csv",
            )
        ),
    ]
    for path in paths:
        assert os.path.realpath(path) in expected_paths
    # valid datetime
    dt = datetime.strptime(
        downloader.cache_dict["test_resource1"]["date_downloaded"],
        "%Y-%m-%d %H:%M:%S.%f",
    )
    assert isinstance(dt, datetime)
    assert isinstance(downloader.cache_dict["test_resource1"]["url"], list)
    assert len(downloader.cache_dict["test_resource1"]["url"]) == 2
    assert downloader.cache_dict["test_resource1"]["lifetime"] == 0
    assert isinstance(downloader.cache_dict["test_resource2"]["url"], list)
    assert len(downloader.cache_dict["test_resource2"]["url"]) == 1
    assert downloader.cache_dict["test_resource2"]["lifetime"] == 0


@pytest.mark.skip(reason="Inconsistent FTP server response")
def test_download_directory_and_caching():
    # use temp dir, no cache file present
    downloader = Downloader(cache_dir=None)
    assert os.path.exists(downloader.cache_dir)
    assert os.path.exists(downloader.cache_file)
    resource = FileDownload(
        "ot_indication",
        "ftp://ftp.ebi.ac.uk/pub/databases/opentargets/platform/24.09/output/etl/parquet/go",
        lifetime=7,
        is_dir=True,
    )
    paths = downloader.download(resource)
    assert len(paths) == 17
    for path in paths:
        assert os.path.exists(path)

    # test caching
    paths = downloader.download(resource)
    # should not download again
    assert len(paths) == 17
    assert "tmp" in paths[0]


def test_download_zip_and_expiration():
    # use temp dir, no cache file present
    downloader = Downloader(cache_dir=None)
    assert os.path.exists(downloader.cache_dir)
    assert os.path.exists(downloader.cache_file)
    resource = FileDownload(
        "test_resource",
        "https://github.com/biocypher/biocypher/raw/main/test/test_CSVs.zip",
        lifetime=7,
    )
    paths = downloader.download(resource)
    with open(downloader.cache_file, "r") as f:
        cache = json.load(f)
    assert cache["test_resource"]["url"][0] == "https://github.com/biocypher/biocypher/raw/main/test/test_CSVs.zip"
    assert cache["test_resource"]["lifetime"] == 7
    assert cache["test_resource"]["date_downloaded"]

    for path in paths:
        assert os.path.exists(path)

    # use files downloaded here and manipulate cache file to test expiration
    downloader.cache_dict["test_resource"]["date_downloaded"] = str(datetime.now() - timedelta(days=4))
    paths = downloader.download(resource)
    # should not download again
    assert "tmp" in paths[0]

    # minus 8 days from date_downloaded
    downloader.cache_dict["test_resource"]["date_downloaded"] = str(datetime.now() - timedelta(days=8))

    paths = downloader.download(resource)
    # should download again
    assert paths[0] is not None


@pytest.mark.parametrize(
    "downloader",
    [
        "downloader_without_specified_cache_dir",
        "downloader_with_specified_cache_dir",
    ],
    indirect=True,
)
def test_cache_api_request(downloader):
    api1 = APIRequest(
        name="uniprot_api",
        url_s=[
            "https://rest.uniprot.org/uniprotkb/P12345.json",
            "https://rest.uniprot.org/uniprotkb/P69905.json",
        ],
        lifetime=1,
    )
    api2 = APIRequest(
        name="intact_api",
        url_s="https://www.ebi.ac.uk/intact/ws/interactor/countTotal",
        lifetime=1,
    )

    paths = downloader.download(api1, api2)

    # test download list
    assert isinstance(paths, list)
    assert len(paths) == 3  # 2 API requests from api1, 1 API request from api2
    for path in paths:
        assert os.path.exists(path)
    assert f"{os.sep}{api1.name}{os.sep}P12345.json" in paths[0]
    assert f"{os.sep}{api1.name}{os.sep}P69905.json" in paths[1]
    assert f"{os.sep}{api2.name}{os.sep}countTotal.json" in paths[2]

    # api1 and api2 have been cached
    # test cached api request
    test_paths = downloader.download(api1, api2)  # get the path(s) of cached API request(s)
    assert isinstance(test_paths, list)
    assert len(paths) == len(test_paths)

    paths.sort()
    test_paths.sort()
    for i in range(len(paths)):
        with open(paths[i], "r") as file1:
            api_request1 = json.load(file1)

        with open(test_paths[i], "r") as file2:
            api_request2 = json.load(file2)
        assert api_request1 == api_request2


def test_api_expiration():
    downloader = Downloader(cache_dir=None)
    assert os.path.exists(downloader.cache_dir)
    assert os.path.exists(downloader.cache_file)
    resource = APIRequest(
        "test_resource",
        "https://rest.uniprot.org/uniprotkb/P12345.json",
        lifetime=7,
    )
    paths = downloader.download(resource)
    with open(downloader.cache_file, "r") as f:
        cache = json.load(f)
    assert cache["test_resource"]["url"][0] == "https://rest.uniprot.org/uniprotkb/P12345.json"
    assert cache["test_resource"]["lifetime"] == 7
    assert cache["test_resource"]["date_downloaded"]

    assert os.path.exists(paths[0])

    # use files downloaded here and manipulate cache file to test expiration
    downloader.cache_dict["test_resource"]["date_downloaded"] = str(datetime.now() - timedelta(days=4))
    paths = downloader.download(resource)
    # should not download again
    assert "tmp" in paths[0]

    # minus 8 days from date_downloaded
    downloader.cache_dict["test_resource"]["date_downloaded"] = str(datetime.now() - timedelta(days=8))

    paths = downloader.download(resource)
    # should download again
    assert paths[0] is not None


def test_download_with_parameter():
    downloader = Downloader(cache_dir=None)
    resource = FileDownload(
        name="zenodo",
        url_s="https://zenodo.org/records/7773985/files/CollecTRI_source.tsv?download=1",
        lifetime=1,
    )

    paths1 = downloader.download(resource)
    assert isinstance(paths1, list)
    assert os.path.exists(paths1[0])
    assert "?download=1" not in paths1[0]

    with open(downloader.cache_file, "r") as f:
        cache = json.load(f)
    assert cache["zenodo"]["url"][0] == "https://zenodo.org/records/7773985/files/CollecTRI_source.tsv?download=1"
    # load resource from cache
    paths2 = downloader.download(resource)
    assert "tmp" in paths2[0]


@patch("requests.get")
def test_download_with_long_url(mock_get):
    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.json = lambda: {"data": "test content"}
    mock_response.raise_for_status = Mock()
    mock_get.return_value = mock_response

    # Create a long URL that would exceed filename length limits
    long_url = "https://example.com/api/data/" + "x" * 200 + "?" + "y" * 200
    downloader = Downloader(cache_dir=None)
    resource = APIRequest(
        name="test_long_url",
        url_s=long_url,
        lifetime=1,
    )
    paths = downloader.download(resource)
    # Verify file path exists
    assert os.path.exists(paths[0])
    # Check filename length is within limits
    filename = os.path.basename(paths[0])
    assert len(filename) <= 150

    # Verify the downloaded file contains our mock data
    with open(paths[0], "r") as f:
        content = f.read()
        assert "test content" in content


@patch("requests.get")
def test_download_file_with_long_url(mock_get):
    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.content = b"test file content"
    mock_response.raise_for_status = Mock()
    mock_response.headers = {"content-length": "15"}  # Length of "test file content"
    # Mock iter_content to return an iterator of our test content
    mock_response.iter_content = lambda chunk_size: iter([b"test file content"])
    mock_get.return_value = mock_response

    # Create a long URL that would exceed filename length limits
    long_url = "https://example.com/files/data/" + "x" * 200 + "?" + "y" * 200
    downloader = Downloader(cache_dir=None)
    resource = FileDownload(
        name="test_long_file_url",
        url_s=long_url,
        lifetime=1,
    )
    paths = downloader.download(resource)
    # Verify file path exists
    assert os.path.exists(paths[0])
    # Check filename length is within limits
    filename = os.path.basename(paths[0])
    assert len(filename) <= 150

    # Verify the downloaded file contains our mock data
    with open(paths[0], "rb") as f:
        content = f.read()
        assert b"test file content" in content



import pytest

from biocypher._create import BioCypherEdge, BioCypherNode


def test_translate_nodes(translator):
    id_type = [
        (
            "G9205",
            "protein",
            {
                "taxon": 9606,
            },
        ),
        (
            "hsa-miR-132-3p",
            "mirna",
            {
                "taxon": 9606,
            },
        ),
        (
            "ASDB_OSBS",
            "complex",
            {
                "taxon": 9606,
            },
        ),
        ("REACT:25520", "reactome", {}),
        ("agpl:001524", "agpl", {}),
    ]
    translated_nodes = translator.translate_nodes(id_type)

    assert all(type(n) is BioCypherNode for n in translated_nodes)

    translated_nodes = translator.translate_nodes(id_type)
    assert next(translated_nodes).get_label() == "protein"
    assert next(translated_nodes).get_label() == "microRNA"
    assert next(translated_nodes).get_label() == "complex"
    assert next(translated_nodes).get_label() == "reactome.pathway"
    assert next(translated_nodes).get_label() == "altered gene product level"


def test_specific_and_generic_ids(translator):
    id_type = [
        (
            "CHAT",
            "hgnc",
            {
                "taxon": 9606,
            },
        ),
        ("REACT:25520", "reactome", {}),
    ]
    translated_nodes = list(translator.translate_nodes(id_type))

    assert translated_nodes[0].get_id() == "CHAT"
    assert translated_nodes[0].get_properties().get("preferred_id") == "hgnc"
    assert translated_nodes[0].get_properties().get("id") == "CHAT"
    assert translated_nodes[1].get_id() == "REACT:25520"
    assert translated_nodes[1].get_properties().get("preferred_id") == "reactome"
    assert translated_nodes[1].get_properties().get("id") == "REACT:25520"


def test_translate_edges(translator):
    # edge type association (defined in `schema_config.yaml`)
    src_tar_type_edge = [
        ("G15258", "MONDO1", "gene_disease", {}),
        ("G15258", "MONDO2", "protein_disease", {}),
        ("G15258", "G15242", "phosphorylation", {}),
    ]

    def gen_edges():
        yield from src_tar_type_edge

    translated_edges = translator.translate_edges(gen_edges())

    assert type(next(translated_edges)) is BioCypherEdge
    assert next(translated_edges).get_label() == "PERTURBED_IN_DISEASE"
    assert next(translated_edges).get_label() == "phosphorylation"

    # node type association (defined in `schema_config.yaml`)
    src_tar_type_node = [
        (
            "G21058",
            "G50127",
            "post_translational",
            {
                "prop1": "test",
            },
        ),
        (
            "G22418",
            "G50123",
            "post_translational",
            {
                "directed": "arbitrary_string",
            },
        ),
        (
            "G15258",
            "G16347",
            "post_translational",
            {
                "directed": True,
                "effect": -1,
            },
        ),
    ]
    translated_edges = translator.translate_edges(src_tar_type_node)
    translated_edges = list(translated_edges)

    node_1 = translated_edges[0]
    node_2 = translated_edges[1]
    node_3 = translated_edges[2]

    assert node_1.get_source_edge().get_label() == "IS_PART_OF"
    assert node_2.get_source_edge().get_label() == "IS_PART_OF"
    assert node_3.get_target_edge().get_label() == "IS_TARGET_OF"
    assert (
        type(node_1.get_node()) is BioCypherNode
        and type(node_1.get_source_edge()) is BioCypherEdge
        and type(node_1.get_target_edge()) is BioCypherEdge
    )
    assert node_3.get_node().get_id() == "G15258_G16347_True_-1"
    assert node_3.get_source_edge().get_source_id() == "G15258"


# def test_biolink_adapter(version_node, translator):
#     # current Biolink model (as opposed to rest of tests)
#     ad = BiolinkAdapter(version_node.extended_schema, translator, clear_cache=True)
#     ver = ad.biolink_version

#     assert isinstance(
#         ad.biolink_leaves['protein']['class_definition'],
#         ClassDefinition,
#     )
#     assert ver

# def test_custom_bmt_yaml(version_node, translator):
#     ad = BiolinkAdapter(
#         version_node.extended_schema,
#         translator,
#         schema=module_data_path('test-biolink-model'),
#         clear_cache=True,
#     )
#     p = ad.biolink_leaves['protein']

#     assert p['class_definition'].description == 'Test'

# def test_biolink_yaml_extension(biolink_adapter):
#     p1 = biolink_adapter.biolink_leaves['post translational interaction']
#     p2 = biolink_adapter.biolink_leaves['phosphorylation']

#     assert (
#         p1['class_definition'].description
#         == 'A pairwise interaction between two proteins' and
#         'PairwiseMolecularInteraction' in p1['ancestors'] and
#         'Entity' in p1['ancestors'] and p2['class_definition'].description
#         == 'The action of one protein phosphorylating another protein' and
#         'PostTranslationalInteraction' in p2['ancestors'] and
#         'Entity' in p2['ancestors']
#     )

# def test_translate_identifiers(translator):
#     # representation of a different schema
#     # host and guest db (which to translate)
#     # TODO
#     pass


def test_merge_multiple_inputs_node(ontology_mapping, translator):
    # Gene has two input labels and one preferred ID
    # no virtual leaves should be created
    # both inputs should lead to creation of the same node type

    # define nodes
    id_type = [
        (
            "CHAT",
            "hgnc",
            {
                "taxon": 9606,
            },
        ),
        (
            "CHRNA4",
            "ensg",
            {
                "taxon": 9606,
            },
        ),
    ]
    translated_nodes = list(translator.translate_nodes(id_type))

    assert translated_nodes

    # check unique node type
    assert not any([schema for schema in ontology_mapping.extended_schema.keys() if ".gene" in schema])
    assert any([schema for schema in ontology_mapping.extended_schema.keys() if ".pathway" in schema])

    # check translator.translate_nodes for unique return type
    assert all([type(node) is BioCypherNode for node in translated_nodes])
    assert all([node.get_label() == "gene" for node in translated_nodes])


def test_implicit_inheritance_node(translator):
    id_type = [
        (
            "snrna1",
            "intact_snrna",
            {},
        ),
        (
            "snrna2",
            "rnacentral_snrna",
            {},
        ),
    ]

    translated_nodes = list(translator.translate_nodes(id_type))

    assert all([type(n) is BioCypherNode for n in translated_nodes])
    assert translated_nodes[0].get_label() == "intact.snRNA sequence"
    assert translated_nodes[1].get_label() == "rnacentral.snRNA sequence"


def test_merge_multiple_inputs_edge(ontology_mapping, translator):
    # GeneToDiseaseAssociation has two input labels and one preferred ID
    # no virtual leaves should be created
    # both inputs should lead to creation of the same edge type
    # intersects with driver/writer function

    # define nodes
    src_tar_type = [
        (
            "CHAT",
            "AD",
            "gene_disease",
            {
                "taxon": 9606,
            },
        ),
        (
            "CHRNA4",
            "AD",
            "protein_disease",
            {
                "taxon": 9606,
            },
        ),
    ]
    translated_edges = list(translator.translate_edges(src_tar_type))

    # check unique edge type
    assert not any(
        [schema for schema in ontology_mapping.extended_schema.keys() if ".gene to disease association" in schema],
    )
    assert any(
        [schema for schema in ontology_mapping.extended_schema.keys() if ".sequence variant" in schema],
    )

    # check translator.translate_nodes for unique return type
    assert all([type(edge) is BioCypherEdge for edge in translated_edges])
    assert all([edge.get_label() == "PERTURBED_IN_DISEASE" for edge in translated_edges])


def test_implicit_inheritance_edge(translator):
    src_tar_type = [
        (
            "mut1",
            "var1",
            "gene1",
            "VARIANT_FOUND_IN_GENE_Known_variant_Gene",
            {},
        ),
        (
            "mut2",
            "var2",
            "gene2",
            "VARIANT_FOUND_IN_GENE_Somatic_mutation_Gene",
            {},
        ),
    ]
    translated_edges = list(translator.translate_edges(src_tar_type))

    assert all([type(edge) is BioCypherEdge for edge in translated_edges])
    assert translated_edges[0].get_label() == "known.sequence variant.variant to gene association"
    assert translated_edges[1].get_label() == "somatic.sequence variant.variant to gene association"


def test_virtual_leaves_inherit_is_a(ontology_mapping):
    snrna = ontology_mapping.extended_schema.get("intact.snRNA sequence")

    assert "is_a" in snrna.keys()
    assert snrna["is_a"] == ["snRNA sequence", "nucleic acid entity"]

    dsdna = ontology_mapping.extended_schema.get("intact.dsDNA sequence")

    assert dsdna["is_a"] == [
        "dsDNA sequence",
        "DNA sequence",
        "nucleic acid entity",
    ]


def test_virtual_leaves_inherit_properties(ontology_mapping):
    snrna = ontology_mapping.extended_schema.get("intact.snRNA sequence")

    assert "properties" in snrna.keys()
    assert "exclude_properties" in snrna.keys()


def test_inherit_properties(ontology_mapping):
    dsdna = ontology_mapping.extended_schema.get("intact.dsDNA sequence")

    assert "properties" in dsdna.keys()
    assert "sequence" in dsdna["properties"]


def test_properties_from_config(translator):
    id_type = [
        (
            "G49205",
            "protein",
            {
                "taxon": 9606,
                "name": "test",
            },
        ),
        (
            "G92035",
            "protein",
            {
                "taxon": 9606,
            },
        ),
        (
            "G92205",
            "protein",
            {
                "taxon": 9606,
                "name": "test2",
                "test": "should_not_be_returned",
            },
        ),
    ]
    translated_edges = translator.translate_nodes(id_type)

    translated_edges_as_list = list(translated_edges)
    assert (
        "name" in translated_edges_as_list[0].get_properties().keys()
        and "name" in translated_edges_as_list[1].get_properties().keys()
        and "test" not in translated_edges_as_list[2].get_properties().keys()
    )

    src_tar_type = [
        (
            "G49205",
            "AD",
            "gene_gene",
            {
                "directional": True,
                "score": 0.5,
                "id": "should_not_be_returned",
            },
        ),
        (
            "G92035",
            "AD",
            "gene_gene",
            {
                "directional": False,
                "curated": True,
                "score": 0.5,
                "test": "should_not_be_returned",
                "id": "should_not_be_returned",
            },
        ),
    ]

    translated_edges = translator.translate_edges(src_tar_type)

    translated_edges_as_list = list(translated_edges)
    assert (
        "directional" in translated_edges_as_list[0].get_properties().keys()
        and "directional" in translated_edges_as_list[1].get_properties().keys()
        and "curated" in translated_edges_as_list[1].get_properties().keys()
        and "score" in translated_edges_as_list[0].get_properties().keys()
        and "score" in translated_edges_as_list[1].get_properties().keys()
        and "test" not in translated_edges_as_list[1].get_properties().keys()
        and "id" not in translated_edges_as_list[0].get_properties().keys()
        and "id" not in translated_edges_as_list[1].get_properties().keys()
    )


def test_exclude_properties(translator):
    id_type = [
        (
            "CHAT",
            "ensg",
            {
                "taxon": 9606,
                "accession": "should_not_be_returned",
            },
        ),
        (
            "ACHE",
            "ensg",
            {
                "taxon": 9606,
            },
        ),
    ]
    translated_nodes = translator.translate_nodes(id_type)

    translated_nodes_as_list = list(translated_nodes)
    assert (
        "taxon" in translated_nodes_as_list[0].get_properties().keys()
        and "taxon" in translated_nodes_as_list[1].get_properties().keys()
        and "accession" not in translated_nodes_as_list[0].get_properties().keys()
    )

    src_tar_type = [
        (
            "G49205",
            "AD",
            "gene_disease",
            {
                "directional": True,
                "score": 0.5,
            },
        ),
        (
            "G92035",
            "AD",
            "gene_disease",
            {
                "directional": False,
                "score": 0.5,
                "accession": "should_not_be_returned",
            },
        ),
    ]

    translated_nodes = translator.translate_edges(src_tar_type)

    translated_nodes_as_list = list(translated_nodes)
    assert (
        "directional" in translated_nodes_as_list[0].get_properties().keys()
        and "directional" in translated_nodes_as_list[1].get_properties().keys()
        and "score" in translated_nodes_as_list[0].get_properties().keys()
        and "score" in translated_nodes_as_list[1].get_properties().keys()
        and "accession" not in translated_nodes_as_list[1].get_properties().keys()
    )


# we need to load the adapter because the mappings are passed from the adapter
# to the translator
def test_translate_term(translator):
    assert translator.translate_term("hgnc") == "Gene"
    assert translator.translate_term("protein_disease") == "PERTURBED_IN_DISEASE"


def test_reverse_translate_term(translator):
    assert "hgnc" in translator.reverse_translate_term("Gene")
    assert "protein_disease" in translator.reverse_translate_term(
        "PERTURBED_IN_DISEASE",
    )


def test_translate_query(translator):
    # we translate to PascalCase for cypher queries, not to internal
    # sentence case
    query = "MATCH (n:hgnc)-[r:gene_disease]->(d:Disease) RETURN n"
    assert translator.translate(query) == "MATCH (n:Gene)-[r:PERTURBED_IN_DISEASE]->(d:Disease) RETURN n"


def test_reverse_translate_query(translator):
    # TODO cannot use sentence case in this context. include sentence to
    # pascal case and back in translation?
    query = "MATCH (n:Known.SequenceVariant)-[r:Known.SequenceVariant.VariantToGeneAssociation]->(g:Gene) RETURN n"
    with pytest.raises(NotImplementedError):
        translator.reverse_translate(query)

    query = "MATCH (n:Known.SequenceVariant)-[r:Known.SequenceVariant.VariantToGeneAssociation]->(g:Protein) RETURN n"
    assert (
        translator.reverse_translate(query)
        == "MATCH (n:Known_variant)-[r:VARIANT_FOUND_IN_GENE_Known_variant_Gene]->(g:protein) RETURN n"
    )


def test_log_missing_nodes(translator):
    translated_nodes = translator.translate_nodes(
        [
            (
                "G49205",
                "missing_protein",
                {
                    "taxon": 9606,
                },
            ),
            ("G92035", "missing_protein", {}),
            ("REACT:25520", "missing_pathway", {}),
        ],
    )

    translated_nodes = list(translated_nodes)

    missing_types = translator.get_missing_biolink_types()
    assert missing_types.get("missing_protein") == 2
    assert missing_types.get("missing_pathway") == 1


def test_strict_mode_error(translator):
    translator.strict_mode = True

    node_1 = (
        "n2",
        "Test",
        {"prop": "val", "source": "test", "licence": "test", "version": "test"},
    )

    assert list(translator.translate_nodes([node_1])) is not None

    # test 'license' instead of 'licence'
    node_2 = (
        "n2",
        "Test",
        {"prop": "val", "source": "test", "license": "test", "version": "test"},
    )

    assert list(translator.translate_nodes([node_2])) is not None

    node_3 = ("n1", "Test", {"prop": "val"})

    with pytest.raises(ValueError):
        list(translator.translate_nodes([node_1, node_2, node_3]))

    edge_1 = (
        "n1",
        "n2",
        "Test",
        {
            "prop": "val",
            "source": "test",
            "licence": "test",
            "version": "test",
        },
    )

    assert list(translator.translate_edges([edge_1])) is not None

    edge_2 = ("n1", "n2", "Test", {"prop": "val"})

    with pytest.raises(ValueError):
        list(translator.translate_edges([edge_1, edge_2]))


def test_strict_mode_property_filter(translator):
    translator.strict_mode = True

    protein_1 = (
        "p1",
        "protein",
        {
            "taxon": 9606,
            "source": "test",
            "licence": "test",
            "version": "test",
        },
    )

    translated_protein_node = list(translator.translate_nodes([protein_1]))

    assert "source" in translated_protein_node[0].get_properties().keys()
    assert "licence" in translated_protein_node[0].get_properties().keys()
    assert "version" in translated_protein_node[0].get_properties().keys()
